@@ -279,7 +279,8 @@ done:
 __dead static void
 usage_init(void)
 {
-	fprintf(stderr, "usage: %s init [-b branch] repository-path\n",
+	fprintf(stderr, "usage: %s init [-A hashing-algorithm] [-b branch]"
+	    " repository-path\n",
 	    getprogname());
 	exit(1);
 }
@@ -290,6 +291,7 @@ cmd_init(int argc, char *argv[])
 	const struct got_error *error = NULL;
 	const char *head_name = NULL;
 	char *repo_path = NULL;
+	enum got_hash_algorithm algo = GOT_HASH_SHA1;
 	int ch;
 
 #ifndef PROFILE
@@ -297,8 +299,17 @@ cmd_init(int argc, char *argv[])
 		err(1, "pledge");
 #endif
 
-	while ((ch = getopt(argc, argv, "b:")) != -1) {
+	while ((ch = getopt(argc, argv, "A:b:")) != -1) {
 		switch (ch) {
+		case 'A':
+			if (!strcmp(optarg, "sha1"))
+				algo = GOT_HASH_SHA1;
+			else if (!strcmp(optarg, "sha256"))
+				algo = GOT_HASH_SHA256;
+			else
+				return got_error_path(optarg,
+				    GOT_ERR_OBJECT_FORMAT);
+			break;
 		case 'b':
 			head_name = optarg;
 			break;
@@ -329,7 +340,7 @@ cmd_init(int argc, char *argv[])
 	if (error)
 		goto done;
 
-	error = got_repo_init(repo_path, head_name);
+	error = got_repo_init(repo_path, head_name, algo);
 done:
 	free(repo_path);
 	return error;
@@ -393,6 +404,9 @@ cmd_info(int argc, char *argv[])
 		goto done;
 
 	printf("repository: %s\n", got_repo_get_path_git_dir(repo));
+	printf("hashing algorithm: %s\n",
+	    got_repo_get_object_format(repo) == GOT_HASH_SHA1 ? "sha1"
+	    : "sha256");
 
 	gotconfig = got_repo_get_gotconfig(repo);
 	if (gotconfig) {
@@ -504,7 +518,7 @@ print_load_info(FILE *out, int print_colored, int print_found, int print_trees,
 static const struct got_error *
 pack_progress(void *arg, int ncolored, int nfound, int ntrees,
     off_t packfile_size, int ncommits, int nobj_total, int nobj_deltify,
-    int nobj_written)
+    int nobj_written, int pack_done)
 {
 	struct got_pack_progress_arg *a = arg;
 	char scaled_size[FMT_SCALED_STRSIZE];
@@ -605,17 +619,18 @@ pack_progress(void *arg, int ncolored, int nfound, int ntrees,
 	if (print_written)
 		fprintf(a->out, "; writing pack: %*s %d%%",
 		    FMT_SCALED_STRSIZE - 2, scaled_size, p_written);
-	if (print_searching || print_total || print_deltify ||
-	    print_written) {
+	if (print_searching || print_total || print_deltify || print_written) {
 		a->printed_something = 1;
 		fflush(a->out);
 	}
+	if (pack_done)
+		fprintf(a->out, "\n");
 	return NULL;
 }
 
 static const struct got_error *
 pack_index_progress(void *arg, off_t packfile_size, int nobj_total,
-    int nobj_indexed, int nobj_loose, int nobj_resolved)
+    int nobj_indexed, int nobj_loose, int nobj_resolved, int indexing_done)
 {
 	struct got_pack_progress_arg *a = arg;
 	char scaled_size[FMT_SCALED_STRSIZE];
@@ -662,7 +677,9 @@ pack_index_progress(void *arg, off_t packfile_size, int nobj_total,
 		printf("; indexing %d%%", p_indexed);
 	if (print_resolved)
 		printf("; resolving deltas %d%%", p_resolved);
-	if (print_size || print_indexed || print_resolved)
+	if (indexing_done)
+		printf("\n");
+	if (print_size || print_indexed || print_resolved || indexing_done)
 		fflush(stdout);
 
 	return NULL;
@@ -703,7 +720,7 @@ cmd_pack(int argc, char *argv[])
 	struct got_repository *repo = NULL;
 	int ch, i, loose_obj_only = 1, force_refdelta = 0, verbosity = 0;
 	struct got_object_id *pack_hash = NULL;
-	char *id_str = NULL;
+	char *id_str = NULL, *idxpath = NULL;
 	struct got_pack_progress_arg ppa;
 	FILE *packfile = NULL;
 	struct got_pathlist_head exclude_args;
@@ -713,7 +730,7 @@ cmd_pack(int argc, char *argv[])
 	struct got_reflist_entry *re, *new;
 	int *pack_fds = NULL;
 
-	TAILQ_INIT(&exclude_args);
+	RB_INIT(&exclude_args);
 	TAILQ_INIT(&exclude_refs);
 	TAILQ_INIT(&include_refs);
 
@@ -743,7 +760,7 @@ cmd_pack(int argc, char *argv[])
 			break;
 		case 'x':
 			got_path_strip_trailing_slashes(optarg);
-			error = got_pathlist_append(&exclude_args,
+			error = got_pathlist_insert(NULL, &exclude_args,
 			    optarg, NULL);
 			if (error)
 				return error;
@@ -773,7 +790,7 @@ cmd_pack(int argc, char *argv[])
 	if (error)
 		goto done;
 
-	TAILQ_FOREACH(pe, &exclude_args, entry) {
+	RB_FOREACH(pe, got_pathlist_head, &exclude_args) {
 		const char *refname = pe->path;
 		error = add_ref(&new, &exclude_refs, refname, repo);
 		if (error)
@@ -828,7 +845,7 @@ cmd_pack(int argc, char *argv[])
 	if (verbosity >= 0)
 		printf("\nWrote %s.pack\n", id_str);
 
-	error = got_repo_index_pack(packfile, pack_hash, repo,
+	error = got_repo_index_pack(&idxpath, packfile, pack_hash, repo,
 	    pack_index_progress, &ppa, check_cancelled, NULL);
 	if (error)
 		goto done;
@@ -849,6 +866,7 @@ done:
 	free(id_str);
 	free(pack_hash);
 	free(repo_path);
+	free(idxpath);
 	return error;
 }
 
@@ -867,7 +885,7 @@ cmd_indexpack(int argc, char *argv[])
 	struct got_repository *repo = NULL;
 	int ch;
 	struct got_object_id *pack_hash = NULL;
-	char *packfile_path = NULL;
+	char *packfile_path = NULL, *idxpath = NULL;
 	char *id_str = NULL;
 	struct got_pack_progress_arg ppa;
 	FILE *packfile = NULL;
@@ -923,7 +941,7 @@ cmd_indexpack(int argc, char *argv[])
 	if (error)
 		goto done;
 
-	error = got_repo_index_pack(packfile, pack_hash, repo,
+	error = got_repo_index_pack(&idxpath, packfile, pack_hash, repo,
 	    pack_index_progress, &ppa, check_cancelled, NULL);
 	if (error)
 		goto done;
@@ -939,6 +957,7 @@ done:
 	}
 	free(id_str);
 	free(pack_hash);
+	free(idxpath);
 	return error;
 }
 
@@ -1250,6 +1269,7 @@ cmd_cleanup(int argc, char *argv[])
 	int ch, dry_run = 0, verbosity = 0;
 	int ncommits = 0, nloose = 0, npacked = 0;
 	int remove_lonely_packidx = 0, ignore_mtime = 0;
+	struct got_pack_progress_arg ppa;
 	struct got_cleanup_progress_arg cpa;
 	struct got_lonely_packidx_progress_arg lpa;
 	off_t loose_before, loose_after;
@@ -1263,7 +1283,7 @@ cmd_cleanup(int argc, char *argv[])
 	int *pack_fds = NULL;
 
 #ifndef PROFILE
-	if (pledge("stdio rpath wpath cpath flock proc exec sendfd unveil",
+	if (pledge("stdio rpath wpath cpath fattr flock proc exec sendfd unveil",
 	    NULL) == -1)
 		err(1, "pledge");
 #endif
@@ -1337,11 +1357,16 @@ cmd_cleanup(int argc, char *argv[])
 	cpa.dry_run = dry_run;
 	cpa.verbosity = verbosity;
 
+	memset(&ppa, 0, sizeof(ppa));
+	ppa.out = stdout;
+	ppa.verbosity = verbosity;
+
 	error = got_repo_cleanup(repo, &loose_before, &loose_after,
 	    &pack_before, &pack_after, &ncommits, &nloose, &npacked,
 	    dry_run, ignore_mtime, cleanup_progress, &cpa,
+	    pack_progress, &ppa, pack_index_progress, &ppa,
 	    check_cancelled, NULL);
-	if (cpa.printed_something)
+	if (ppa.printed_something || cpa.printed_something)
 		printf("\n");
 	if (error)
 		goto done;
@@ -1420,7 +1445,7 @@ cmd_dump(int argc, char *argv[])
 	int verbosity = 0;
 	int i, ch;
 
-	TAILQ_INIT(&exclude_args);
+	RB_INIT(&exclude_args);
 	TAILQ_INIT(&exclude_refs);
 	TAILQ_INIT(&include_refs);
 
@@ -1443,7 +1468,7 @@ cmd_dump(int argc, char *argv[])
 			got_path_strip_trailing_slashes(repo_path);
 			break;
 		case 'x':
-			error = got_pathlist_append(&exclude_args,
+			error = got_pathlist_insert(NULL, &exclude_args,
 			    optarg, NULL);
 			if (error)
 				return error;
@@ -1472,7 +1497,7 @@ cmd_dump(int argc, char *argv[])
 	if (error)
 		goto done;
 
-	TAILQ_FOREACH(pe, &exclude_args, entry) {
+	RB_FOREACH(pe, got_pathlist_head, &exclude_args) {
 		refname = pe->path;
 		error = add_ref(&new, &exclude_refs, refname, repo);
 		if (error)
@@ -1546,7 +1571,7 @@ load_progress(void *arg, off_t packfile_size, int nobj_total,
     int nobj_indexed, int nobj_loose, int nobj_resolved)
 {
 	return pack_index_progress(arg, packfile_size, nobj_total,
-	    nobj_indexed, nobj_loose, nobj_resolved);
+	    nobj_indexed, nobj_loose, nobj_resolved, 0);
 }
 
 static int
@@ -1554,10 +1579,10 @@ is_wanted_ref(struct got_pathlist_head *wanted, const char *ref)
 {
 	struct got_pathlist_entry *pe;
 
-	if (TAILQ_EMPTY(wanted))
+	if (RB_EMPTY(wanted))
 		return 1;
 
-	TAILQ_FOREACH(pe, wanted, entry) {
+	RB_FOREACH(pe, got_pathlist_head, wanted) {
 		if (strcmp(pe->path, ref) == 0)
 			return 1;
 	}
@@ -1670,8 +1695,8 @@ cmd_load(int argc, char *argv[])
 	int verbosity = 0;
 	int ch, i;
 
-	TAILQ_INIT(&include_args);
-	TAILQ_INIT(&available_refs);
+	RB_INIT(&include_args);
+	RB_INIT(&available_refs);
 
 #ifndef PROFILE
 	if (pledge("stdio rpath wpath cpath fattr flock proc exec "
@@ -1718,7 +1743,7 @@ cmd_load(int argc, char *argv[])
 		got_path_strip_trailing_slashes(refname);
 		if (!got_ref_name_is_valid(refname))
 			errx(1, "invalid reference name %s", refname);
-		error = got_pathlist_append(&include_args, refname, NULL);
+		error = got_pathlist_insert(NULL, &include_args, refname, NULL);
 		if (error)
 			goto done;
 	}
@@ -1751,7 +1776,7 @@ cmd_load(int argc, char *argv[])
 		goto done;
 
 	if (list_refs_only) {
-		TAILQ_FOREACH(pe, &available_refs, entry) {
+		RB_FOREACH(pe, got_pathlist_head, &available_refs) {
 			const char *refname = pe->path;
 			struct got_object_id *id = pe->data;
 			char *idstr;
@@ -1770,7 +1795,7 @@ cmd_load(int argc, char *argv[])
 		goto done;
 
 	/* Update references */
-	TAILQ_FOREACH(pe, &available_refs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, &available_refs) {
 		const struct got_error *unlock_err;
 		struct got_reference *ref;
 		const char *refname = pe->path;
