@@ -20,6 +20,7 @@
 #define GOTD_UNIX_SOCKET_BACKLOG 10
 #define GOTD_USER	"_gotd"
 #define GOTD_CONF_PATH	"/etc/gotd.conf"
+#define GOTD_SECRETS_PATH "/etc/gotd-secrets.conf"
 #ifndef GOTD_EMPTY_PATH
 #define GOTD_EMPTY_PATH	"/var/empty"
 #endif
@@ -113,8 +114,8 @@ struct gotd_notification_target {
 			char *hostname;
 			char *port;
 			char *path;
-			char *user;
-			char *password;
+			char *auth;
+			char *hmac;
 		} http;
 	} conf;
 };
@@ -155,6 +156,7 @@ struct gotd_uid_connection_limit {
 
 struct gotd_child_proc;
 
+struct gotd_secrets;
 struct gotd {
 	pid_t pid;
 	char unix_socket_path[PATH_MAX];
@@ -168,6 +170,7 @@ struct gotd {
 	struct timeval auth_timeout;
 	struct gotd_uid_connection_limit *connection_limits;
 	size_t nconnection_limits;
+	struct gotd_secrets *secrets;
 
 	char *argv0;
 	const char *confpath;
@@ -176,7 +179,7 @@ struct gotd {
 };
 
 enum gotd_imsg_type {
-	/* An error occured while processing a request. */
+	/* An error occurred while processing a request. */
 	GOTD_IMSG_ERROR,
 
 	/* Commands used by gotctl(8). */
@@ -223,7 +226,7 @@ enum gotd_imsg_type {
 	GOTD_IMSG_REF_UPDATES_START, /* Ref updates starting. */
 	GOTD_IMSG_REF_UPDATE_OK, /* Update went OK. */
 	GOTD_IMSG_REF_UPDATE_NG, /* Update was not good. */
-	GOTD_IMSG_REFS_UPDATED, /* The server proccessed all ref updates. */
+	GOTD_IMSG_REFS_UPDATED, /* The server processed all ref updates. */
 
 	/* Client connections. */
 	GOTD_IMSG_DISCONNECT,
@@ -242,7 +245,11 @@ enum gotd_imsg_type {
 	GOTD_IMSG_CONNECT_NOTIFIER,
 	GOTD_IMSG_CONNECT_SESSION,
 	GOTD_IMSG_NOTIFY,
-	GOTD_IMSG_NOTIFICATION_SENT
+	GOTD_IMSG_NOTIFICATION_SENT,
+
+	/* Secrets. */
+	GOTD_IMSG_SECRETS,	/* number of secrets */
+	GOTD_IMSG_SECRET,
 };
 
 /* Structure for GOTD_IMSG_ERROR. */
@@ -485,7 +492,8 @@ struct gotd_imsg_notify {
 };
 
 int enter_chroot(const char *);
-int parse_config(const char *, enum gotd_procid, struct gotd *);
+int parse_config(const char *, enum gotd_procid, struct gotd_secrets *,
+    struct gotd *);
 struct gotd_repo *gotd_find_repo_by_name(const char *, struct gotd_repolist *);
 struct gotd_repo *gotd_find_repo_by_path(const char *, struct gotd *);
 struct gotd_uid_connection_limit *gotd_find_uid_connection_limit(
@@ -496,7 +504,6 @@ const struct got_error *gotd_parse_url(char **, char **, char **,
 
 /* imsg.c */
 const struct got_error *gotd_imsg_flush(struct imsgbuf *);
-const struct got_error *gotd_imsg_recv(struct imsg *, struct imsgbuf *, size_t);
 const struct got_error *gotd_imsg_poll_recv(struct imsg *, struct imsgbuf *,
     size_t);
 const struct got_error *gotd_imsg_recv_error(uint32_t *client_id,
