@@ -508,7 +508,7 @@ struct repo_read_pack_progress_arg {
 static const struct got_error *
 pack_progress(void *arg, int ncolored, int nfound, int ntrees,
     off_t packfile_size, int ncommits, int nobj_total, int nobj_deltify,
-    int nobj_written)
+    int nobj_written, int pack_done)
 {
 	struct repo_read_pack_progress_arg *a = arg;
 	struct gotd_imsg_packfile_progress iprog;
@@ -687,7 +687,7 @@ repo_read_dispatch_session(int fd, short event, void *arg)
 
 	if (event & EV_READ) {
 		if ((n = imsgbuf_read(ibuf)) == -1)
-			fatal("imsg_read error");
+			fatal("imsgbuf_read error");
 		if (n == 0)	/* Connection closed. */
 			shut = 1;
 	}
@@ -808,7 +808,7 @@ repo_read_dispatch(int fd, short event, void *arg)
 
 	if (event & EV_READ) {
 		if ((n = imsgbuf_read(ibuf)) == -1)
-			fatal("imsg_read error");
+			fatal("imsgbuf_read error");
 		if (n == 0)	/* Connection closed. */
 			shut = 1;
 	}
