@@ -171,7 +171,7 @@ send_peeled_tag_ref(struct got_reference *ref, struct got_object *obj,
 	}
 
 	/* Keep in sync with struct gotd_imsg_ref definition. */
-	if (imsg_add(wbuf, id->sha1, SHA1_DIGEST_LENGTH) == -1) {
+	if (imsg_add(wbuf, id->hash, SHA1_DIGEST_LENGTH) == -1) {
 		err = got_error_from_errno("imsg_add REF");
 		goto done;
 	}
@@ -219,7 +219,7 @@ send_ref(struct got_reference *ref, struct imsgbuf *ibuf)
 	}
 
 	/* Keep in sync with struct gotd_imsg_ref definition. */
-	if (imsg_add(wbuf, id->sha1, SHA1_DIGEST_LENGTH) == -1)
+	if (imsg_add(wbuf, id->hash, SHA1_DIGEST_LENGTH) == -1)
 		return got_error_from_errno("imsg_add REF");
 	if (imsg_add(wbuf, &namelen, sizeof(namelen)) == -1)
 		return got_error_from_errno("imsg_add REF");
@@ -250,14 +250,9 @@ list_refs(struct imsg *imsg)
 	size_t datalen;
 	struct gotd_imsg_reflist irefs;
 	struct imsgbuf ibuf;
-	int client_fd;
 
 	TAILQ_INIT(&refs);
 
-	client_fd = imsg_get_fd(imsg);
-	if (client_fd == -1)
-		return got_error(GOT_ERR_PRIVSEP_NO_FD);
-
 	datalen = imsg->hdr.len - IMSG_HEADER_SIZE;
 	if (datalen != 0)
 		return got_error(GOT_ERR_PRIVSEP_LEN);
@@ -268,13 +263,18 @@ list_refs(struct imsg *imsg)
 	}
 	repo_write.refs_listed = 1;
 
-	client->fd = client_fd;
+	client->fd = imsg_get_fd(imsg);
+	if (client->fd == -1)
+		return got_error(GOT_ERR_PRIVSEP_NO_FD);
+
 	client->nref_updates = 0;
 	client->nref_del = 0;
 	client->nref_new = 0;
 	client->nref_move = 0;
 
-	imsg_init(&ibuf, client_fd);
+	if (imsgbuf_init(&ibuf, client->fd) == -1)
+		return got_error_from_errno("imsgbuf_init");
+	imsgbuf_allow_fdpass(&ibuf);
 
 	err = got_ref_list(&refs, repo_write.repo, "",
 	    got_ref_cmp_by_name, NULL);
@@ -320,7 +320,7 @@ list_refs(struct imsg *imsg)
 	err = gotd_imsg_flush(&ibuf);
 done:
 	got_ref_list_free(&refs);
-	imsg_clear(&ibuf);
+	imsgbuf_clear(&ibuf);
 	return err;
 }
 
@@ -365,7 +365,7 @@ verify_object_type(struct got_object_id *id, int expected_obj_type,
 
 	idx = got_packidx_get_object_idx(packidx, id);
 	if (idx == -1) {
-		got_sha1_digest_to_str(id->sha1, hex, sizeof(hex));
+		got_object_id_hex(id, hex, sizeof(hex));
 		return got_error_fmt(GOT_ERR_BAD_PACKFILE,
 		    "object %s is missing from pack file", hex);
 	}
@@ -376,7 +376,7 @@ verify_object_type(struct got_object_id *id, int expected_obj_type,
 		return err;
 
 	if (obj->type != expected_obj_type) {
-		got_sha1_digest_to_str(id->sha1, hex, sizeof(hex));
+		got_object_id_hex(id, hex, sizeof(hex));
 		got_object_type_label(&typestr, expected_obj_type);
 		err = got_error_fmt(GOT_ERR_OBJ_TYPE,
 		    "%s is not pointing at a %s object", hex, typestr);
@@ -435,7 +435,7 @@ protect_require_yca(struct got_object_id *tip_id,
 		goto done;
 
 	if (obj_type != GOT_OBJ_TYPE_COMMIT) {
-		got_sha1_digest_to_str(expected_yca_id->sha1, hex, sizeof(hex));
+		got_object_id_hex(expected_yca_id, hex, sizeof(hex));
 		err = got_error_fmt(GOT_ERR_OBJ_TYPE,
 		    "%s is not pointing at a commit object", hex);
 		goto done;
@@ -490,8 +490,7 @@ protect_require_yca(struct got_object_id *tip_id,
 
 			idx = got_packidx_get_object_idx(packidx, &qid->id);
 			if (idx == -1) {
-				got_sha1_digest_to_str(qid->id.sha1,
-				    hex, sizeof(hex));
+				got_object_id_hex(&qid->id, hex, sizeof(hex));
 				err = got_error_fmt(GOT_ERR_BAD_PACKFILE,
 				    "object %s is missing from pack file", hex);
 				goto done;
@@ -503,8 +502,7 @@ protect_require_yca(struct got_object_id *tip_id,
 				goto done;
 
 			if (obj->type != GOT_OBJ_TYPE_COMMIT) {
-				got_sha1_digest_to_str(qid->id.sha1,
-				    hex, sizeof(hex));
+				got_object_id_hex(&qid->id, hex, sizeof(hex));
 				err = got_error_fmt(GOT_ERR_OBJ_TYPE,
 				    "%s is not pointing at a commit object",
 				    hex);
@@ -516,7 +514,8 @@ protect_require_yca(struct got_object_id *tip_id,
 			if (err)
 				goto done;
 
-			err = got_object_parse_commit(&commit, buf, len);
+			err = got_object_parse_commit(&commit, buf, len,
+			    GOT_HASH_SHA1);
 			if (err)
 				goto done;
 
@@ -630,11 +629,15 @@ recv_ref_update(struct imsg *imsg)
 	if (datalen != sizeof(iref) + iref.name_len)
 		return got_error(GOT_ERR_PRIVSEP_LEN);
 
-	imsg_init(&ibuf, client->fd);
+	if (imsgbuf_init(&ibuf, client->fd))
+		return got_error_from_errno("imsgbuf_init");
+	imsgbuf_allow_fdpass(&ibuf);
 
 	refname = strndup(imsg->data + sizeof(iref), iref.name_len);
-	if (refname == NULL)
-		return got_error_from_errno("strndup");
+	if (refname == NULL) {
+		err = got_error_from_errno("strndup");
+		goto done;
+	}
 
 	ref_update = calloc(1, sizeof(*ref_update));
 	if (ref_update == NULL) {
@@ -642,14 +645,14 @@ recv_ref_update(struct imsg *imsg)
 		goto done;
 	}
 
-	memcpy(ref_update->old_id.sha1, iref.old_id, SHA1_DIGEST_LENGTH);
-	memcpy(ref_update->new_id.sha1, iref.new_id, SHA1_DIGEST_LENGTH);
+	memcpy(ref_update->old_id.hash, iref.old_id, SHA1_DIGEST_LENGTH);
+	memcpy(ref_update->new_id.hash, iref.new_id, SHA1_DIGEST_LENGTH);
 
 	err = got_ref_open(&ref, repo_write.repo, refname, 0);
 	if (err) {
 		if (err->code != GOT_ERR_NOT_REF)
 			goto done;
-		if (memcmp(ref_update->new_id.sha1,
+		if (memcmp(ref_update->new_id.hash,
 		    zero_id, sizeof(zero_id)) == 0) {
 			err = got_error_fmt(GOT_ERR_BAD_OBJ_ID,
 			    "%s", refname);
@@ -705,7 +708,7 @@ recv_ref_update(struct imsg *imsg)
 	    repo_write.pid);
 
 	ref_update->ref = ref;
-	if (memcmp(ref_update->new_id.sha1, zero_id, sizeof(zero_id)) == 0) {
+	if (memcmp(ref_update->new_id.hash, zero_id, sizeof(zero_id)) == 0) {
 		ref_update->delete_ref = 1;
 		client->nref_del++;
 	}
@@ -835,8 +838,8 @@ copy_ref_delta(int infd, int outfd, off_t *outsize, BUF *buf, size_t *buf_pos,
 }
 
 static const struct got_error *
-copy_offset_delta(int infd, int outfd, off_t *outsize, BUF *buf, size_t *buf_pos,
-    struct got_hash *ctx)
+copy_offset_delta(int infd, int outfd, off_t *outsize, BUF *buf,
+    size_t *buf_pos, struct got_hash *ctx)
 {
 	const struct got_error *err = NULL;
 	uint64_t o = 0;
@@ -1164,7 +1167,9 @@ report_pack_status(const struct got_error *unpack_err)
 	const char *unpack_ok = "unpack ok\n";
 	size_t len;
 
-	imsg_init(&ibuf, client->fd);
+	if (imsgbuf_init(&ibuf, client->fd))
+		return got_error_from_errno("imsgbuf_init");
+	imsgbuf_allow_fdpass(&ibuf);
 
 	if (unpack_err)
 		istatus.reason_len = strlen(unpack_err->msg);
@@ -1194,7 +1199,7 @@ report_pack_status(const struct got_error *unpack_err)
 
 	err = gotd_imsg_flush(&ibuf);
 done:
-	imsg_clear(&ibuf);
+	imsgbuf_clear(&ibuf);
 	return err;
 }
 
@@ -1204,6 +1209,7 @@ recv_packfile(int *have_packfile, struct imsg *imsg)
 	const struct got_error *err = NULL, *unpack_err;
 	struct repo_write_client *client = &repo_write_client;
 	struct gotd_imsg_recv_packfile ireq;
+	struct got_object_id id;
 	FILE *tempfiles[3] = { NULL, NULL, NULL };
 	struct repo_tempfile {
 		int fd;
@@ -1211,7 +1217,6 @@ recv_packfile(int *have_packfile, struct imsg *imsg)
 	} repo_tempfiles[3] = { { - 1, - 1 }, { - 1, - 1 }, { - 1, - 1 }, };
 	int i;
 	size_t datalen;
-	struct imsgbuf ibuf;
 	struct got_ratelimit rl;
 	struct got_pack *pack = NULL;
 	off_t pack_filesize = 0;
@@ -1230,17 +1235,17 @@ recv_packfile(int *have_packfile, struct imsg *imsg)
 	if (client->pack_pipe == -1 || client->packidx_fd == -1)
 		return got_error(GOT_ERR_PRIVSEP_NO_FD);
 
-	imsg_init(&ibuf, client->fd);
-
 	pack = &client->pack;
 	memset(pack, 0, sizeof(*pack));
 	pack->fd = imsg_get_fd(imsg);
-	if (pack->fd == -1)
-		return got_error(GOT_ERR_PRIVSEP_NO_FD);
+	if (pack->fd == -1) {
+		err = got_error(GOT_ERR_PRIVSEP_NO_FD);
+		goto done;
+	}
 
 	err = got_delta_cache_alloc(&pack->delta_cache);
 	if (err)
-		return err;
+		goto done;
 
 	for (i = 0; i < nitems(repo_tempfiles); i++) {
 		struct repo_tempfile *t = &repo_tempfiles[i];
@@ -1267,10 +1272,6 @@ recv_packfile(int *have_packfile, struct imsg *imsg)
 		tempfiles[i] = f;
 	}
 
-	err = gotd_imsg_flush(&ibuf);
-	if (err)
-		goto done;
-
 	log_debug("receiving pack data");
 	unpack_err = recv_packdata(&pack_filesize, &nobj,
 	    client->pack_sha1, client->pack_pipe, pack->fd);
@@ -1321,10 +1322,14 @@ recv_packfile(int *have_packfile, struct imsg *imsg)
 	pack->filesize = pack_filesize;
 	*have_packfile = 1;
 
+	memset(&id, 0, sizeof(id));
+	memcpy(&id.hash, client->pack_sha1, SHA1_DIGEST_LENGTH);
+	id.algo = GOT_HASH_SHA1;
+
 	log_debug("begin indexing pack (%lld bytes in size)",
 	    (long long)pack->filesize);
 	err = got_pack_index(pack, client->packidx_fd,
-	    tempfiles[0], tempfiles[1], tempfiles[2], client->pack_sha1,
+	    tempfiles[0], tempfiles[1], tempfiles[2], &id,
 	    pack_index_progress, NULL, &rl);
 	if (err)
 		goto done;
@@ -1351,7 +1356,6 @@ done:
 	}
 	if (err)
 		got_pack_close(pack);
-	imsg_clear(&ibuf);
 	return err;
 }
 
@@ -1399,7 +1403,8 @@ verify_packfile(void)
 		if (ref_update->delete_ref)
 			continue;
 
-		TAILQ_FOREACH(pe, repo_write.protected_tag_namespaces, entry) {
+		RB_FOREACH(pe, got_pathlist_head,
+		    repo_write.protected_tag_namespaces) {
 			err = protect_tag_namespace(pe->path, &client->pack,
 			    packidx, ref_update);
 			if (err)
@@ -1422,7 +1427,7 @@ verify_packfile(void)
 			int idx = got_packidx_get_object_idx(packidx,
 			    &ref_update->new_id);
 			if (idx == -1) {
-				got_sha1_digest_to_str(ref_update->new_id.sha1,
+				got_object_id_hex(&ref_update->new_id,
 				    hex, sizeof(hex));
 				err = got_error_fmt(GOT_ERR_BAD_PACKFILE,
 				    "object %s is missing from pack file",
@@ -1431,14 +1436,15 @@ verify_packfile(void)
 			}
 		}
 
-		TAILQ_FOREACH(pe, repo_write.protected_branch_namespaces,
-		    entry) {
+		RB_FOREACH(pe, got_pathlist_head,
+		    repo_write.protected_branch_namespaces) {
 			err = protect_branch_namespace(pe->path,
 			    &client->pack, packidx, ref_update);
 			if (err)
 				goto done;
 		}
-		TAILQ_FOREACH(pe, repo_write.protected_branches, entry) {
+		RB_FOREACH(pe, got_pathlist_head,
+		    repo_write.protected_branches) {
 			err = protect_branch(pe->path, &client->pack,
 			    packidx, ref_update);
 			if (err)
@@ -1471,20 +1477,67 @@ protect_refs_from_deletion(void)
 
 		refname = got_ref_get_name(ref_update->ref);
 
-		TAILQ_FOREACH(pe, repo_write.protected_tag_namespaces, entry) {
+		RB_FOREACH(pe, got_pathlist_head,
+		    repo_write.protected_tag_namespaces) {
 			err = protect_ref_namespace(refname, pe->path);
 			if (err)
 				return err;
 		}
 
-		TAILQ_FOREACH(pe, repo_write.protected_branch_namespaces,
-		    entry) {
+		RB_FOREACH(pe, got_pathlist_head,
+		    repo_write.protected_branch_namespaces) {
 			err = protect_ref_namespace(refname, pe->path);
 			if (err)
 				return err;
 		}
 
-		TAILQ_FOREACH(pe, repo_write.protected_branches, entry) {
+		RB_FOREACH(pe, got_pathlist_head,
+		    repo_write.protected_branches) {
+			if (strcmp(refname, pe->path) == 0) {
+				return got_error_fmt(GOT_ERR_REF_PROTECTED,
+				    "%s", refname);
+			}
+		}
+	}
+
+	return NULL;
+}
+
+static const struct got_error *
+protect_refs_from_moving(void)
+{
+	const struct got_error *err = NULL;
+	struct repo_write_client *client = &repo_write_client;
+	struct gotd_ref_update *ref_update;
+	struct got_pathlist_entry *pe;
+	const char *refname;
+
+	STAILQ_FOREACH(ref_update, &client->ref_updates, entry) {
+		if (ref_update->delete_ref)
+			continue;
+
+		if (got_object_id_cmp(&ref_update->old_id,
+		    &ref_update->new_id) == 0)
+			continue;
+
+		refname = got_ref_get_name(ref_update->ref);
+
+		RB_FOREACH(pe, got_pathlist_head,
+		    repo_write.protected_tag_namespaces) {
+			err = protect_ref_namespace(refname, pe->path);
+			if (err)
+				return err;
+		}
+
+		RB_FOREACH(pe, got_pathlist_head,
+		    repo_write.protected_branch_namespaces) {
+			err = protect_ref_namespace(refname, pe->path);
+			if (err)
+				return err;
+		}
+
+		RB_FOREACH(pe, got_pathlist_head,
+		    repo_write.protected_branches) {
 			if (strcmp(refname, pe->path) == 0) {
 				return got_error_fmt(GOT_ERR_REF_PROTECTED,
 				    "%s", refname);
@@ -1540,8 +1593,8 @@ send_ref_update(struct gotd_ref_update *ref_update, struct gotd_imsgev *iev)
 	size_t len;
 
 	memset(&iref, 0, sizeof(iref));
-	memcpy(iref.old_id, ref_update->old_id.sha1, SHA1_DIGEST_LENGTH);
-	memcpy(iref.new_id, ref_update->new_id.sha1, SHA1_DIGEST_LENGTH);
+	memcpy(iref.old_id, ref_update->old_id.hash, SHA1_DIGEST_LENGTH);
+	memcpy(iref.new_id, ref_update->new_id.hash, SHA1_DIGEST_LENGTH);
 	iref.ref_is_new = ref_update->ref_is_new;
 	iref.delete_ref = ref_update->delete_ref;
 	iref.name_len = strlen(refname);
@@ -1628,17 +1681,13 @@ receive_pack_idx(struct imsg *imsg, struct gotd_imsgev *iev)
 }
 
 static const struct got_error *
-notify_removed_ref(const char *refname, uint8_t *sha1,
+notify_removed_ref(const char *refname, struct got_object_id *id,
     struct gotd_imsgev *iev, int fd)
 {
 	const struct got_error *err;
-	struct got_object_id id;
 	char *id_str;
 
-	memset(&id, 0, sizeof(id));
-	memcpy(id.sha1, sha1, sizeof(id.sha1));
-
-	err = got_object_id_str(&id_str, &id);
+	err = got_object_id_str(&id_str, id);
 	if (err)
 		return err;
 
@@ -1719,13 +1768,13 @@ print_diffstat(struct got_diffstat_cb_arg *dsa, int fd)
 {
 	struct got_pathlist_entry *pe;
 
-	TAILQ_FOREACH(pe, dsa->paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, dsa->paths) {
 		struct got_diff_changed_path *cp = pe->data;
 		int pad = dsa->max_path_len - pe->path_len + 1;
 
 		dprintf(fd, " %c  %s%*c | %*d+ %*d-\n", cp->status,
-		     pe->path, pad, ' ', dsa->add_cols + 1, cp->add,
-		     dsa->rm_cols + 1, cp->rm);
+		    pe->path, pad, ' ', dsa->add_cols + 1, cp->add,
+		    dsa->rm_cols + 1, cp->rm);
 	}
 	dprintf(fd,
 	    "\n%d file%s changed, %d insertion%s(+), %d deletion%s(-)\n\n",
@@ -1882,7 +1931,7 @@ print_commits(struct got_object_id *root_id, struct got_object_id *end_id,
 	const int shortlog_threshold = 50;
 
 	STAILQ_INIT(&reversed_commits);
-	TAILQ_INIT(&changed_paths);
+	RB_INIT(&changed_paths);
 
 	/* XXX first-parent only for now */
 	err = got_commit_graph_open(&graph, "/", 1);
@@ -1950,11 +1999,7 @@ print_commits(struct got_object_id *root_id, struct got_object_id *end_id,
 done:
 	if (commit)
 		got_object_commit_close(commit);
-	while (!STAILQ_EMPTY(&reversed_commits)) {
-		qid = STAILQ_FIRST(&reversed_commits);
-		STAILQ_REMOVE_HEAD(&reversed_commits, entry);
-		got_object_qid_free(qid);
-	}
+	got_object_id_queue_free(&reversed_commits);
 	got_pathlist_free(&changed_paths, GOT_PATHLIST_FREE_ALL);
 	got_commit_graph_close(graph);
 	return err;
@@ -2025,42 +2070,36 @@ done:
 }
 
 static const struct got_error *
-notify_changed_ref(const char *refname, uint8_t *old_sha1,
-    uint8_t *new_sha1, struct gotd_imsgev *iev, int fd)
+notify_changed_ref(const char *refname, struct got_object_id *old_id,
+    struct got_object_id *new_id, struct gotd_imsgev *iev, int fd)
 {
 	const struct got_error *err;
-	struct got_object_id old_id, new_id;
 	int old_obj_type, new_obj_type;
 	const char *label;
 	char *new_id_str = NULL;
 
-	memset(&old_id, 0, sizeof(old_id));
-	memcpy(old_id.sha1, old_sha1, sizeof(old_id.sha1));
-	memset(&new_id, 0, sizeof(new_id));
-	memcpy(new_id.sha1, new_sha1, sizeof(new_id.sha1));
-
-	err = got_object_get_type(&old_obj_type, repo_write.repo, &old_id);
+	err = got_object_get_type(&old_obj_type, repo_write.repo, old_id);
 	if (err)
 		return err;
 
-	err = got_object_get_type(&new_obj_type, repo_write.repo, &new_id);
+	err = got_object_get_type(&new_obj_type, repo_write.repo, new_id);
 	if (err)
 		return err;
 
 	switch (new_obj_type) {
 	case GOT_OBJ_TYPE_COMMIT:
-		err = print_commits(&new_id,
-		    old_obj_type == GOT_OBJ_TYPE_COMMIT ? &old_id : NULL,
+		err = print_commits(new_id,
+		    old_obj_type == GOT_OBJ_TYPE_COMMIT ? old_id : NULL,
 		    repo_write.repo, fd);
 		break;
 	case GOT_OBJ_TYPE_TAG:
-		err = print_tag(&new_id, refname, repo_write.repo, fd);
+		err = print_tag(new_id, refname, repo_write.repo, fd);
 		break;
 	default:
 		err = got_object_type_label(&label, new_obj_type);
 		if (err)
 			goto done;
-		err = got_object_id_str(&new_id_str, &new_id);
+		err = got_object_id_str(&new_id_str, new_id);
 		if (err)
 			goto done;
 		dprintf(fd, "%s: %s object %s\n", refname, label, new_id_str);
@@ -2072,24 +2111,20 @@ done:
 }
 
 static const struct got_error *
-notify_created_ref(const char *refname, uint8_t *sha1,
+notify_created_ref(const char *refname, struct got_object_id *id,
     struct gotd_imsgev *iev, int fd)
 {
 	const struct got_error *err;
-	struct got_object_id id;
 	int obj_type;
 
-	memset(&id, 0, sizeof(id));
-	memcpy(id.sha1, sha1, sizeof(id.sha1));
-
-	err = got_object_get_type(&obj_type, repo_write.repo, &id);
+	err = got_object_get_type(&obj_type, repo_write.repo, id);
 	if (err)
 		return err;
 
 	if (obj_type == GOT_OBJ_TYPE_TAG)
-		return print_tag(&id, refname, repo_write.repo, fd);
+		return print_tag(id, refname, repo_write.repo, fd);
 
-	return print_commits(&id, NULL, repo_write.repo, fd);
+	return print_commits(id, NULL, repo_write.repo, fd);
 }
 
 static const struct got_error *
@@ -2127,13 +2162,13 @@ render_notification(struct imsg *imsg, struct gotd_imsgev *iev)
 
 	switch (ireq.action) {
 	case GOTD_NOTIF_ACTION_CREATED:
-		err = notify_created_ref(refname, ireq.new_id, iev, fd);
+		err = notify_created_ref(refname, &ireq.new_id, iev, fd);
 		break;
 	case GOTD_NOTIF_ACTION_REMOVED:
-		err = notify_removed_ref(refname, ireq.old_id, iev, fd);
+		err = notify_removed_ref(refname, &ireq.old_id, iev, fd);
 		break;
 	case GOTD_NOTIF_ACTION_CHANGED:
-		err = notify_changed_ref(refname, ireq.old_id, ireq.new_id,
+		err = notify_changed_ref(refname, &ireq.old_id, &ireq.new_id,
 		    iev, fd);
 		break;
 	}
@@ -2182,18 +2217,16 @@ repo_write_dispatch_session(int fd, short event, void *arg)
 	int shut = 0, have_packfile = 0;
 
 	if (event & EV_READ) {
-		if ((n = imsg_read(ibuf)) == -1 && errno != EAGAIN)
-			fatal("imsg_read error");
+		if ((n = imsgbuf_read(ibuf)) == -1)
+			fatal("imsgbuf_read error");
 		if (n == 0)	/* Connection closed. */
 			shut = 1;
 	}
 
 	if (event & EV_WRITE) {
-		n = msgbuf_write(&ibuf->w);
-		if (n == -1 && errno != EAGAIN)
-			fatal("msgbuf_write");
-		if (n == 0)	/* Connection closed. */
-			shut = 1;
+		err = gotd_imsg_flush(ibuf);
+		if (err)
+			fatalx("%s", err->msg);
 	}
 
 	for (;;) {
@@ -2262,6 +2295,14 @@ repo_write_dispatch_session(int fd, short event, void *arg)
 				 */
 				repo_write.repo->pack_path_mtime.tv_sec = 0;
 				repo_write.repo->pack_path_mtime.tv_nsec = 0;
+			} else {
+				/*
+				 * Clients sending empty pack files might be
+				 * attempting to move a protected reference.
+				 */
+				err = protect_refs_from_moving();
+				if (err)
+					break;
 			}
 			err = update_refs(iev);
 			if (err) {
@@ -2284,9 +2325,8 @@ repo_write_dispatch_session(int fd, short event, void *arg)
 	}
 
 	if (!shut && check_cancelled(NULL) == NULL) {
-		if (err &&
-		    gotd_imsg_send_error_event(iev, PROC_REPO_WRITE,
-		        client->id, err) == -1) {
+		if (err && gotd_imsg_send_error_event(iev, PROC_REPO_WRITE,
+		    client->id, err) == -1) {
 			log_warnx("could not send error to parent: %s",
 			    err->msg);
 		}
@@ -2315,7 +2355,9 @@ recv_connect(struct imsg *imsg)
 	if (repo_write.session_fd == -1)
 		return got_error(GOT_ERR_PRIVSEP_NO_FD);
 
-	imsg_init(&iev->ibuf, repo_write.session_fd);
+	if (imsgbuf_init(&iev->ibuf, repo_write.session_fd) == -1)
+		return got_error_from_errno("imsgbuf_init");
+	imsgbuf_allow_fdpass(&iev->ibuf);
 	iev->handler = repo_write_dispatch_session;
 	iev->events = EV_READ;
 	iev->handler_arg = NULL;
@@ -2338,21 +2380,24 @@ repo_write_dispatch(int fd, short event, void *arg)
 	struct repo_write_client *client = &repo_write_client;
 
 	if (event & EV_READ) {
-		if ((n = imsg_read(ibuf)) == -1 && errno != EAGAIN)
-			fatal("imsg_read error");
-		if (n == 0)	/* Connection closed. */
+		if ((n = imsgbuf_read(ibuf)) == -1)
+			fatal("imsgbuf_read error");
+		if (n == 0) {	/* Connection closed. */
 			shut = 1;
+			goto done;
+		}
 	}
 
 	if (event & EV_WRITE) {
-		n = msgbuf_write(&ibuf->w);
-		if (n == -1 && errno != EAGAIN)
-			fatal("msgbuf_write");
-		if (n == 0)	/* Connection closed. */
-			shut = 1;
+		err = gotd_imsg_flush(ibuf);
+		if (err)
+			fatalx("%s", err->msg);
 	}
 
-	while (err == NULL && check_cancelled(NULL) == NULL) {
+	while (err == NULL) {
+		err = check_cancelled(NULL);
+		if (err)
+			break;
 		if ((n = imsg_get(ibuf, &imsg)) == -1)
 			fatal("%s: imsg_get", __func__);
 		if (n == 0)	/* No more messages. */
@@ -2364,19 +2409,18 @@ repo_write_dispatch(int fd, short event, void *arg)
 			break;
 		default:
 			log_debug("unexpected imsg %d", imsg.hdr.type);
+			err = got_error(GOT_ERR_PRIVSEP_MSG);
 			break;
 		}
 
 		imsg_free(&imsg);
 	}
 
-	if (!shut && check_cancelled(NULL) == NULL) {
-		if (err &&
-		    gotd_imsg_send_error_event(iev, PROC_REPO_WRITE,
-		        client->id, err) == -1) {
-			log_warnx("could not send error to parent: %s",
-			    err->msg);
-		}
+	if (err && gotd_imsg_send_error_event(iev, PROC_REPO_WRITE,
+	    client->id, err) == -1)
+		log_warnx("could not send error to parent: %s", err->msg);
+done:
+	if (!shut) {
 		gotd_imsg_event_add(iev);
 	} else {
 		/* This pipe is dead. Remove its event handler */
@@ -2426,6 +2470,11 @@ repo_write_main(const char *title, const char *repo_path,
 		    "bare git repository required");
 		goto done;
 	}
+	if (got_repo_get_object_format(repo_write.repo) != GOT_HASH_SHA1) {
+		err = got_error_msg(GOT_ERR_NOT_IMPL,
+		    "sha256 object IDs unsupported in network protocol");
+		goto done;
+	}
 
 	got_repo_temp_fds_set(repo_write.repo, temp_fds);
 
@@ -2434,7 +2483,11 @@ repo_write_main(const char *title, const char *repo_path,
 	signal(SIGPIPE, SIG_IGN);
 	signal(SIGHUP, SIG_IGN);
 
-	imsg_init(&iev.ibuf, GOTD_FILENO_MSG_PIPE);
+	if (imsgbuf_init(&iev.ibuf, GOTD_FILENO_MSG_PIPE) == -1) {
+		err = got_error_from_errno("imsgbuf_init");
+		goto done;
+	}
+	imsgbuf_allow_fdpass(&iev.ibuf);
 	iev.handler = repo_write_dispatch;
 	iev.events = EV_READ;
 	iev.handler_arg = NULL;
