@@ -72,6 +72,7 @@ static struct gotd_session_read {
 	struct timeval request_timeout;
 	enum gotd_session_read_state state;
 	struct gotd_imsgev repo_child_iev;
+	int repo_child_packfd;
 } gotd_session;
 
 static struct gotd_session_client {
@@ -104,7 +105,7 @@ disconnect(struct gotd_session_client *client)
 	    GOTD_IMSG_DISCONNECT, PROC_SESSION_READ, -1, NULL, 0) == -1)
 		log_warn("imsg compose DISCONNECT");
 
-	imsg_clear(&gotd_session.repo_child_iev.ibuf);
+	imsgbuf_clear(&gotd_session.repo_child_iev.ibuf);
 	event_del(&gotd_session.repo_child_iev.ev);
 	evtimer_del(&client->tmo);
 	close(client->fd);
@@ -133,9 +134,12 @@ disconnect_on_error(struct gotd_session_client *client,
 
 	if (err->code != GOT_ERR_EOF) {
 		log_warnx("uid %d: %s", client->euid, err->msg);
-		imsg_init(&ibuf, client->fd);
-		gotd_imsg_send_error(&ibuf, 0, PROC_SESSION_READ, err);
-		imsg_clear(&ibuf);
+		if (imsgbuf_init(&ibuf, client->fd) == -1) {
+			log_warn("imsgbuf_init");
+		} else {
+			gotd_imsg_send_error(&ibuf, 0, PROC_SESSION_READ, err);
+			imsgbuf_clear(&ibuf);
+		}
 	}
 
 	disconnect(client);
@@ -192,6 +196,7 @@ recv_packfile_done(struct imsg *imsg)
 static void
 session_dispatch_repo_child(int fd, short event, void *arg)
 {
+	const struct got_error *err = NULL;
 	struct gotd_imsgev *iev = arg;
 	struct imsgbuf *ibuf = &iev->ibuf;
 	struct gotd_session_client *client = &gotd_session_client;
@@ -200,8 +205,8 @@ session_dispatch_repo_child(int fd, short event, void *arg)
 	struct imsg imsg;
 
 	if (event & EV_READ) {
-		if ((n = imsg_read(ibuf)) == -1 && errno != EAGAIN)
-			fatal("imsg_read error");
+		if ((n = imsgbuf_read(ibuf)) == -1)
+			fatal("imsgbuf_read error");
 		if (n == 0) {
 			/* Connection closed. */
 			shut = 1;
@@ -210,14 +215,9 @@ session_dispatch_repo_child(int fd, short event, void *arg)
 	}
 
 	if (event & EV_WRITE) {
-		n = msgbuf_write(&ibuf->w);
-		if (n == -1 && errno != EAGAIN)
-			fatal("msgbuf_write");
-		if (n == 0) {
-			/* Connection closed. */
-			shut = 1;
-			goto done;
-		}
+		err = gotd_imsg_flush(ibuf);
+		if (err)
+			fatalx("%s", err->msg);
 	}
 
 	for (;;) {
@@ -431,20 +431,24 @@ send_packfile(struct gotd_session_client *client)
 		return err;
 	}
 
-	/* Send pack pipe end 0 to repo child process. */
-	if (gotd_imsg_compose_event(&gotd_session.repo_child_iev,
+	/*
+	 * Send pack data pipe end 0 to gotsh(1) (expects just an fd, no data).
+	 *
+	 * We will forward the other pipe end to the repo_read process only
+	 * once we have confirmation that gotsh(1) has received its end.
+	 * It is important that we send a pipe end to gotsh(1) before the
+	 * repo_read process starts sending pack progress messages via imsg
+	 * to prevent spurious "unexpected privsep message" errors from gotsh.
+	 */
+	if (gotd_imsg_compose_event(&client->iev,
 	    GOTD_IMSG_PACKFILE_PIPE, PROC_GOTD, pipe[0], NULL, 0) == -1) {
 		err = got_error_from_errno("imsg compose PACKFILE_PIPE");
 		close(pipe[1]);
 		return err;
 	}
 
-	/* Send pack pipe end 1 to gotsh(1) (expects just an fd, no data). */
-	if (gotd_imsg_compose_event(&client->iev,
-	    GOTD_IMSG_PACKFILE_PIPE, PROC_GOTD, pipe[1], NULL, 0) == -1)
-		err = got_error_from_errno("imsg compose PACKFILE_PIPE");
-
-	return err;
+	gotd_session.repo_child_packfd = pipe[1];
+	return NULL;
 }
 
 static void
@@ -458,19 +462,10 @@ session_dispatch_client(int fd, short events, void *arg)
 	ssize_t n;
 
 	if (events & EV_WRITE) {
-		while (ibuf->w.queued) {
-			n = msgbuf_write(&ibuf->w);
-			if (n == -1 && errno != EAGAIN) {
-				err = got_error_from_errno("imsg_flush");
-				disconnect_on_error(client, err);
-				return;
-			}
-			if (n == 0) {
-				/* Connection closed. */
-				err = got_error(GOT_ERR_EOF);
-				disconnect_on_error(client, err);
-				return;
-			}
+		err = gotd_imsg_flush(ibuf);
+		if (err) {
+			disconnect_on_error(client, err);
+			return;
 		}
 
 		if (client->flush_disconnect) {
@@ -479,30 +474,28 @@ session_dispatch_client(int fd, short events, void *arg)
 		}
 	}
 
-	if ((events & EV_READ) == 0)
-		return;
-
-	memset(&imsg, 0, sizeof(imsg));
+	if (events & EV_READ) {
+		n = imsgbuf_read(ibuf);
+		if (n == -1) {
+			err = got_error_from_errno("imsgbuf_read");
+			disconnect_on_error(client, err);
+			return;
+		}
+		if (n == 0) {
+			 err = got_error(GOT_ERR_EOF);
+			 disconnect_on_error(client, err);
+			 return;
+		}
+	}
 
 	while (err == NULL) {
-		err = gotd_imsg_recv(&imsg, ibuf, 0);
-		if (err) {
-			if (err->code == GOT_ERR_PRIVSEP_READ)
-				err = NULL;
-			else if (err->code == GOT_ERR_EOF &&
-			    gotd_session.state ==
-			    GOTD_STATE_EXPECT_CAPABILITIES) {
-				/*
-				 * The client has closed its socket before
-				 * sending its capability announcement.
-				 * This can happen when Git clients have
-				 * no ref-updates to send.
-				 */
-				disconnect_on_error(client, err);
-				return;
-			}
+		n = imsg_get(ibuf, &imsg);
+		if (n == -1) {
+			err = got_error_from_errno("imsg_get");
 			break;
 		}
+		if (n == 0)
+			break;
 
 		evtimer_del(&client->tmo);
 
@@ -557,7 +550,7 @@ session_dispatch_client(int fd, short events, void *arg)
 			client->accept_flush_pkt = 1;
 			break;
 		case GOTD_IMSG_FLUSH:
-			if (gotd_session.state != GOTD_STATE_EXPECT_WANT && 
+			if (gotd_session.state != GOTD_STATE_EXPECT_WANT &&
 			    gotd_session.state !=
 			    GOTD_STATE_EXPECT_HAVE_OR_DONE) {
 				err = got_error_msg(GOT_ERR_BAD_REQUEST,
@@ -609,6 +602,24 @@ session_dispatch_client(int fd, short events, void *arg)
 			client->accept_flush_pkt = 1;
 			err = send_packfile(client);
 			break;
+		case GOTD_IMSG_PACKFILE_READY:
+			if (gotd_session.repo_child_packfd == -1) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			/*
+			 * gotsh(1) has received its send of the pack pipe.
+			 * Send pack pipe end 1 to repo child process.
+			 */
+			if (gotd_imsg_compose_event(
+			    &gotd_session.repo_child_iev,
+			    GOTD_IMSG_PACKFILE_PIPE, PROC_GOTD,
+			    gotd_session.repo_child_packfd, NULL, 0) == -1) {
+				err = got_error_from_errno("imsg compose "
+				    "PACKFILE_PIPE");
+			} else
+				gotd_session.repo_child_packfd = -1;
+			break;
 		default:
 			log_debug("unexpected imsg %d", imsg.hdr.type);
 			err = got_error(GOT_ERR_PRIVSEP_MSG);
@@ -683,7 +694,9 @@ recv_connect(struct imsg *imsg)
 	if (client->username == NULL)
 		return got_error_from_errno("strndup");
 
-	imsg_init(&client->iev.ibuf, client->fd);
+	if (imsgbuf_init(&client->iev.ibuf, client->fd) == -1)
+		return got_error_from_errno("imsgbuf_init");
+	imsgbuf_allow_fdpass(&client->iev.ibuf);
 	client->iev.handler = session_dispatch_client;
 	client->iev.events = EV_READ;
 	client->iev.handler_arg = NULL;
@@ -725,7 +738,11 @@ recv_repo_child(struct imsg *imsg)
 	if (fd == -1)
 		return got_error(GOT_ERR_PRIVSEP_NO_FD);
 
-	imsg_init(&gotd_session.repo_child_iev.ibuf, fd);
+	if (imsgbuf_init(&gotd_session.repo_child_iev.ibuf, fd)) {
+		close(fd);
+		return got_error_from_errno("imsgbuf_init");
+	}
+	imsgbuf_allow_fdpass(&gotd_session.repo_child_iev.ibuf);
 	gotd_session.repo_child_iev.handler = session_dispatch_repo_child;
 	gotd_session.repo_child_iev.events = EV_READ;
 	gotd_session.repo_child_iev.handler_arg = NULL;
@@ -744,6 +761,7 @@ recv_repo_child(struct imsg *imsg)
 static void
 session_dispatch(int fd, short event, void *arg)
 {
+	const struct got_error *err = NULL;
 	struct gotd_imsgev *iev = arg;
 	struct imsgbuf *ibuf = &iev->ibuf;
 	struct gotd_session_client *client = &gotd_session_client;
@@ -752,8 +770,8 @@ session_dispatch(int fd, short event, void *arg)
 	struct imsg imsg;
 
 	if (event & EV_READ) {
-		if ((n = imsg_read(ibuf)) == -1 && errno != EAGAIN)
-			fatal("imsg_read error");
+		if ((n = imsgbuf_read(ibuf)) == -1)
+			fatal("imsgbuf_read error");
 		if (n == 0) {
 			/* Connection closed. */
 			shut = 1;
@@ -762,14 +780,9 @@ session_dispatch(int fd, short event, void *arg)
 	}
 
 	if (event & EV_WRITE) {
-		n = msgbuf_write(&ibuf->w);
-		if (n == -1 && errno != EAGAIN)
-			fatal("msgbuf_write");
-		if (n == 0) {
-			/* Connection closed. */
-			shut = 1;
-			goto done;
-		}
+		err = gotd_imsg_flush(ibuf);
+		if (err)
+			fatalx("%s", err->msg);
 	}
 
 	for (;;) {
@@ -841,8 +854,13 @@ session_read_main(const char *title, const char *repo_path,
 	memcpy(&gotd_session.request_timeout, request_timeout,
 	    sizeof(gotd_session.request_timeout));
 	gotd_session.repo_cfg = repo_cfg;
+	gotd_session.repo_child_packfd = -1;
 
-	imsg_init(&gotd_session.notifier_iev.ibuf, -1);
+	if (imsgbuf_init(&gotd_session.notifier_iev.ibuf, -1) == -1) {
+		err = got_error_from_errno("imsgbuf_init");
+		goto done;
+	}
+	imsgbuf_allow_fdpass(&gotd_session.notifier_iev.ibuf);
 
 	err = got_repo_open(&gotd_session.repo, repo_path, NULL, pack_fds);
 	if (err)
@@ -852,6 +870,11 @@ session_read_main(const char *title, const char *repo_path,
 		    "bare git repository required");
 		goto done;
 	}
+	if (got_repo_get_object_format(gotd_session.repo) != GOT_HASH_SHA1) {
+		err = got_error_msg(GOT_ERR_NOT_IMPL,
+		    "sha256 object IDs unsupported in network protocol");
+		goto done;
+	}
 
 	got_repo_temp_fds_set(gotd_session.repo, temp_fds);
 
@@ -873,7 +896,12 @@ session_read_main(const char *title, const char *repo_path,
 	gotd_session_client.delta_cache_fd = -1;
 	gotd_session_client.accept_flush_pkt = 1;
 
-	imsg_init(&gotd_session.parent_iev.ibuf, GOTD_FILENO_MSG_PIPE);
+	if (imsgbuf_init(&gotd_session.parent_iev.ibuf, GOTD_FILENO_MSG_PIPE)
+	    == -1) {
+		err = got_error_from_errno("imsgbuf_init");
+		goto done;
+	}
+	imsgbuf_allow_fdpass(&gotd_session.parent_iev.ibuf);
 	gotd_session.parent_iev.handler = session_dispatch;
 	gotd_session.parent_iev.events = EV_READ;
 	gotd_session.parent_iev.handler_arg = NULL;
@@ -903,5 +931,7 @@ session_read_shutdown(void)
 	got_repo_pack_fds_close(gotd_session.pack_fds);
 	got_repo_temp_fds_close(gotd_session.temp_fds);
 	free(gotd_session_client.username);
+	if (gotd_session.repo_child_packfd != -1)
+		close(gotd_session.repo_child_packfd);
 	exit(0);
 }
