@@ -47,6 +47,7 @@ void	 fcgi_parse_begin_request(uint8_t *, uint16_t, struct request *,
 void	 fcgi_parse_params(uint8_t *, uint16_t, struct request *, uint16_t);
 int	 fcgi_send_response(struct request *, int, const void *, size_t);
 
+void	 dump_fcgi_request_body(const char *, struct fcgi_record_header *);
 void	 dump_fcgi_record_header(const char *, struct fcgi_record_header *);
 void	 dump_fcgi_begin_request_body(const char *,
 	    struct fcgi_begin_request_body *);
@@ -64,7 +65,7 @@ fcgi_request(int fd, short events, void *arg)
 	size_t parsed = 0;
 
 	n = read(fd, c->buf + c->buf_pos + c->buf_len,
-	    FCGI_RECORD_SIZE - c->buf_pos-c->buf_len);
+	    FCGI_RECORD_SIZE - c->buf_pos - c->buf_len);
 
 	switch (n) {
 	case -1:
@@ -78,7 +79,7 @@ fcgi_request(int fd, short events, void *arg)
 		break;
 
 	case 0:
-		log_debug("closed connection");
+		log_info("closed connection");
 		goto fail;
 	default:
 		break;
@@ -101,7 +102,7 @@ fcgi_request(int fd, short events, void *arg)
 
 		/* drop the parsed record */
 		if (parsed != 0 && c->buf_len > 0) {
-			bcopy(c->buf + c->buf_pos, c->buf, c->buf_len);
+			memmove(c->buf, c->buf + c->buf_pos, c->buf_len);
 			c->buf_pos = 0;
 		}
 	} while (parsed > 0 && c->buf_len > 0);
@@ -117,15 +118,17 @@ fcgi_parse_record(uint8_t *buf, size_t n, struct request *c)
 	struct fcgi_record_header *h;
 
 	if (n < sizeof(struct fcgi_record_header))
-		 return 0;
+		return 0;
 
 	h = (struct fcgi_record_header*) buf;
 
-	dump_fcgi_record("", h);
+	dump_fcgi_record_header("", h);
 
 	if (n < sizeof(struct fcgi_record_header) + ntohs(h->content_len)
 	    + h->padding_len)
-		 return 0;
+		return 0;
+
+	dump_fcgi_request_body("", h);
 
 	if (h->version != 1)
 		log_warn("wrong version");
@@ -308,7 +311,7 @@ send_response(struct request *c, int type, const uint8_t *data,
 	iov[2].iov_base = (void *)padding;
 	iov[2].iov_len = header.padding_len;
 
-	dump_fcgi_record("resp ", &header);
+	dump_fcgi_record_header("resp ", &header);
 
 	/*
 	 * XXX: add some simple write heuristics here
@@ -328,14 +331,13 @@ send_response(struct request *c, int type, const uint8_t *data,
 				nanosleep(&ts, NULL);
 				continue;
 			}
-			log_debug("%s: write failure: %s", __func__,
-			    strerror(errno));
+			log_warn("%s: write failure", __func__);
 			c->sock->client_status = CLIENT_DISCONNECT;
 			return -1;
 		}
 
 		if (nw != tot)
-			log_debug("%s: partial write: %zu vs %zu", __func__,
+			log_warnx("%s: partial write: %zu vs %zu", __func__,
 			    nw, tot);
 
 		tot -= nw;
@@ -413,10 +415,8 @@ fcgi_cleanup_request(struct request *c)
 }
 
 void
-dump_fcgi_record(const char *p, struct fcgi_record_header *h)
+dump_fcgi_request_body(const char *p, struct fcgi_record_header *h)
 {
-	dump_fcgi_record_header(p, h);
-
 	if (h->type == FCGI_BEGIN_REQUEST)
 		dump_fcgi_begin_request_body(p,
 		    (struct fcgi_begin_request_body *)(h + 1));
