@@ -35,6 +35,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <vis.h>
 
 #include "got_error.h"
 #include "got_object.h"
@@ -87,6 +88,8 @@ static const struct got_error *gotweb_assign_querystring(struct querystring *,
 static int gotweb_render_index(struct template *);
 static const struct got_error *gotweb_load_got_path(struct repo_dir **,
     const char *, struct request *);
+static const struct got_error *gotweb_load_file(char **, const char *,
+    const char *, int);
 static const struct got_error *gotweb_get_repo_description(char **,
     struct server *, const char *, int);
 static const struct got_error *gotweb_get_clone_url(char **, struct server *,
@@ -144,6 +147,7 @@ gotweb_process_request(struct request *c)
 	struct server *srv = NULL;
 	struct querystring *qs = NULL;
 	struct repo_dir *repo_dir = NULL;
+	struct repo_commit *commit;
 	const char *rss_ctype = "application/rss+xml;charset=utf-8";
 	const uint8_t *buf;
 	size_t len;
@@ -175,6 +179,38 @@ gotweb_process_request(struct request *c)
 		goto err;
 	}
 
+	/* Log the request. */
+	if (gotwebd_env->gotwebd_verbose > 0) {
+		char *server_name = NULL;
+		char *querystring = NULL;
+		char *document_uri = NULL;
+
+		if (c->server_name[0] &&
+		    stravis(&server_name, c->server_name, VIS_SAFE) == -1) {
+			log_warn("stravis");
+			server_name = NULL;
+		}
+		if (c->querystring[0] &&
+		    stravis(&querystring, c->querystring, VIS_SAFE) == -1) {
+			log_warn("stravis");
+			querystring = NULL;
+		}
+		if (c->document_uri[0] &&
+		    stravis(&document_uri, c->document_uri, VIS_SAFE) == -1) {
+			log_warn("stravis");
+			document_uri = NULL;
+		}
+
+		log_info("processing request: server='%s' query='%s' "
+		    "document_uri='%s'",
+		    server_name ? server_name : "",
+		    querystring ? querystring : "",
+		    document_uri ? document_uri : "");
+		free(server_name);
+		free(querystring);
+		free(document_uri);
+	}
+
 	/*
 	 * certain actions require a commit id in the querystring. this stops
 	 * bad actors from exploiting this by manually manipulating the
@@ -191,9 +227,14 @@ gotweb_process_request(struct request *c)
 	}
 
 	if (qs->action != INDEX) {
+		if (qs->path == NULL) {
+			error = got_error(GOT_ERR_BAD_QUERYSTRING);
+			goto err;
+		}
+
 		error = gotweb_load_got_path(&repo_dir, qs->path, c);
 		c->t->repo_dir = repo_dir;
-		if (error && error->code != GOT_ERR_LONELY_PACKIDX)
+		if (error)
 			goto err;
 	}
 
@@ -358,6 +399,15 @@ gotweb_process_request(struct request *c)
 			goto err;
 		}
 		qs->action = SUMMARY;
+		commit = TAILQ_FIRST(&c->t->repo_commits);
+		if (commit && qs->commit == NULL) {
+			qs->commit = strdup(commit->commit_id);
+			if (qs->commit == NULL) {
+				error = got_error_from_errno("strdup");
+				log_warn("%s: strdup", __func__);
+				goto err;
+			}
+		}
 		if (gotweb_reply(c, 200, "text/html", NULL) == -1)
 			return;
 		gotweb_render_page(c->tp, gotweb_render_summary);
@@ -368,7 +418,7 @@ gotweb_process_request(struct request *c)
 			log_warnx("%s: %s", __func__, error->msg);
 			goto err;
 		}
-		if (c->t->tag_count == 0) {
+		if (TAILQ_EMPTY(&c->t->repo_tags)) {
 			error = got_error_msg(GOT_ERR_BAD_OBJ_ID,
 			    "bad commit id");
 			goto err;
@@ -812,7 +862,7 @@ gotweb_render_index(struct template *tp)
 
 		error = gotweb_load_got_path(&repo_dir, sd_dent[d_i]->d_name,
 		    c);
-		if (error && error->code != GOT_ERR_LONELY_PACKIDX) {
+		if (error) {
 			if (error->code != GOT_ERR_NOT_GIT_REPO)
 				log_warnx("%s: %s: %s", __func__,
 				    sd_dent[d_i]->d_name, error->msg);
@@ -1108,9 +1158,17 @@ gotweb_load_got_path(struct repo_dir **rp, const char *dir,
 	    repo_dir->path, dirfd(dt));
 	if (error)
 		goto err;
-	error = got_get_repo_owner(&repo_dir->owner, c);
-	if (error)
-		goto err;
+	if (srv->show_repo_owner) {
+		error = gotweb_load_file(&repo_dir->owner, repo_dir->path,
+		    "owner", dirfd(dt));
+		if (error)
+			goto err;
+		if (repo_dir->owner == NULL) {
+			error = got_get_repo_owner(&repo_dir->owner, c);
+			if (error)
+				goto err;
+		}
+	}
 	if (srv->show_repo_age) {
 		error = got_get_repo_age(&repo_dir->age, c, NULL);
 		if (error)
@@ -1130,30 +1188,24 @@ err:
 }
 
 static const struct got_error *
-gotweb_get_repo_description(char **description, struct server *srv,
-    const char *dirpath, int dir)
+gotweb_load_file(char **str, const char *dir, const char *file, int dirfd)
 {
 	const struct got_error *error = NULL;
 	struct stat sb;
-	int fd = -1;
 	off_t len;
+	int fd;
 
-	*description = NULL;
-	if (srv->show_repo_description == 0)
-		return NULL;
+	*str = NULL;
 
-	fd = openat(dir, "description", O_RDONLY);
+	fd = openat(dirfd, file, O_RDONLY);
 	if (fd == -1) {
-		if (errno != ENOENT && errno != EACCES) {
-			error = got_error_from_errno_fmt("openat %s/%s",
-			    dirpath, "description");
-		}
-		goto done;
+		if (errno == ENOENT || errno == EACCES)
+			return NULL;
+		return got_error_from_errno_fmt("openat %s/%s", dir, file);
 	}
 
 	if (fstat(fd, &sb) == -1) {
-		error = got_error_from_errno_fmt("fstat %s/%s",
-		    dirpath, "description");
+		error = got_error_from_errno_fmt("fstat %s/%s", dir, file);
 		goto done;
 	}
 
@@ -1161,64 +1213,40 @@ gotweb_get_repo_description(char **description, struct server *srv,
 	if (len > GOTWEBD_MAXDESCRSZ - 1)
 		len = GOTWEBD_MAXDESCRSZ - 1;
 
-	*description = calloc(len + 1, sizeof(**description));
-	if (*description == NULL) {
+	*str = calloc(len + 1, 1);
+	if (*str == NULL) {
 		error = got_error_from_errno("calloc");
 		goto done;
 	}
 
-	if (read(fd, *description, len) == -1)
+	if (read(fd, *str, len) == -1)
 		error = got_error_from_errno("read");
-done:
+ done:
 	if (fd != -1 && close(fd) == -1 && error == NULL)
 		error = got_error_from_errno("close");
 	return error;
 }
 
+static const struct got_error *
+gotweb_get_repo_description(char **description, struct server *srv,
+    const char *dirpath, int dir)
+{
+	*description = NULL;
+	if (srv->show_repo_description == 0)
+		return NULL;
+
+	return gotweb_load_file(description, dirpath, "description", dir);
+}
+
 static const struct got_error *
 gotweb_get_clone_url(char **url, struct server *srv, const char *dirpath,
     int dir)
 {
-	const struct got_error *error = NULL;
-	struct stat sb;
-	int fd = -1;
-	off_t len;
-
 	*url = NULL;
 	if (srv->show_repo_cloneurl == 0)
 		return NULL;
 
-	fd = openat(dir, "cloneurl", O_RDONLY);
-	if (fd == -1) {
-		if (errno != ENOENT && errno != EACCES) {
-			error = got_error_from_errno_fmt("openat %s/%s",
-			    dirpath, "cloneurl");
-		}
-		goto done;
-	}
-
-	if (fstat(fd, &sb) == -1) {
-		error = got_error_from_errno_fmt("fstat %s/%s",
-		    dirpath, "cloneurl");
-		goto done;
-	}
-
-	len = sb.st_size;
-	if (len > GOTWEBD_MAXCLONEURLSZ - 1)
-		len = GOTWEBD_MAXCLONEURLSZ - 1;
-
-	*url = calloc(len + 1, sizeof(**url));
-	if (*url == NULL) {
-		error = got_error_from_errno("calloc");
-		goto done;
-	}
-
-	if (read(fd, *url, len) == -1)
-		error = got_error_from_errno("read");
-done:
-	if (fd != -1 && close(fd) == -1 && error == NULL)
-		error = got_error_from_errno("close");
-	return error;
+	return gotweb_load_file(url, dirpath, "cloneurl", dir);
 }
 
 int
