@@ -60,12 +60,12 @@ void
 imsg_event_add(struct imsgev *iev)
 {
 	if (iev->handler == NULL) {
-		imsg_flush(&iev->ibuf);
+		imsgbuf_flush(&iev->ibuf);
 		return;
 	}
 
 	iev->events = EV_READ;
-	if (iev->ibuf.w.queued)
+	if (imsgbuf_queuelen(&iev->ibuf))
 		iev->events |= EV_WRITE;
 
 	event_del(&iev->ev);
@@ -105,10 +105,7 @@ main_compose_sockets(struct gotwebd *env, uint32_t type, int fd,
 
 		/* prevent fd exhaustion */
 		if (d != -1) {
-			do {
-				ret = imsg_flush(&env->iev_server[i].ibuf);
-			} while (ret == -1 && errno == EAGAIN);
-			if (ret == -1)
+			if (imsgbuf_flush(&env->iev_server[i].ibuf) == -1)
 				goto err;
 			imsg_event_add(&env->iev_server[i]);
 		}
@@ -144,16 +141,14 @@ gotwebd_dispatch_sockets(int fd, short event, void *arg)
 	ibuf = &iev->ibuf;
 
 	if (event & EV_READ) {
-		if ((n = imsg_read(ibuf)) == -1 && errno != EAGAIN)
-			fatal("imsg_read error");
+		if ((n = imsgbuf_read(ibuf)) == -1)
+			fatal("imsgbuf_read error");
 		if (n == 0)	/* Connection closed */
 			shut = 1;
 	}
 	if (event & EV_WRITE) {
-		if ((n = msgbuf_write(&ibuf->w)) == -1 && errno != EAGAIN)
-			fatal("msgbuf_write");
-		if (n == 0)	/* Connection closed */
-			shut = 1;
+		if (imsgbuf_write(ibuf) == -1)
+			fatal("imsgbuf_write");
 	}
 
 	for (;;) {
@@ -210,7 +205,7 @@ gotwebd_sighdlr(int sig, short event, void *arg)
 static int
 spawn_socket_process(struct gotwebd *env, const char *argv0, int n)
 {
-	const char	*argv[6];
+	const char	*argv[8];
 	int		 argc = 0;
 	int		 p[2];
 	pid_t		 pid;
@@ -225,7 +220,9 @@ spawn_socket_process(struct gotwebd *env, const char *argv0, int n)
 		break;
 	default:	/* parent */
 		close(p[0]);
-		imsg_init(&env->iev_server[n].ibuf, p[1]);
+		if (imsgbuf_init(&env->iev_server[n].ibuf, p[1]) == -1)
+			fatal("imsgbuf_init");
+		imsgbuf_allow_fdpass(&env->iev_server[n].ibuf);
 		env->iev_server[n].handler = gotwebd_dispatch_sockets;
 		env->iev_server[n].data = &env->iev_server[n];
 		event_set(&env->iev_server[n].ev, p[1], EV_READ,
@@ -238,6 +235,10 @@ spawn_socket_process(struct gotwebd *env, const char *argv0, int n)
 
 	argv[argc++] = argv0;
 	argv[argc++] = "-S";
+	if (strcmp(env->gotwebd_conffile, GOTWEBD_CONF) != 0) {
+		argv[argc++] = "-f";
+		argv[argc++] = env->gotwebd_conffile;
+	}
 	if (env->gotwebd_debug)
 		argv[argc++] = "-d";
 	if (env->gotwebd_verbose > 0)
@@ -269,6 +270,7 @@ int
 main(int argc, char **argv)
 {
 	struct event		 sigint, sigterm, sighup, sigpipe, sigusr1;
+	struct event_base	*evb;
 	struct gotwebd		*env;
 	struct passwd		*pw;
 	int			 ch, i;
@@ -347,8 +349,8 @@ main(int argc, char **argv)
 	log_setverbose(env->gotwebd_verbose);
 
 	if (server_proc) {
-		setproctitle("sockets");
-		log_procinit("sockets");
+		setproctitle("server");
+		log_procinit("server");
 
 		if (chroot(env->httpd_chroot) == -1)
 			fatal("chroot %s", env->httpd_chroot);
@@ -366,7 +368,7 @@ main(int argc, char **argv)
 	if (!env->gotwebd_debug && daemon(1, 0) == -1)
 		fatal("daemon");
 
-	event_init();
+	evb = event_init();
 
 	env->nserver = env->prefork_gotwebd;
 	env->iev_server = calloc(env->nserver, sizeof(*env->iev_server));
@@ -400,14 +402,16 @@ main(int argc, char **argv)
 	if (gotwebd_configure(env) == -1)
 		fatalx("configuration failed");
 
+	if (setgroups(1, &pw->pw_gid) == -1 ||
+	    setresgid(pw->pw_gid, pw->pw_gid, pw->pw_gid) == -1 ||
+	    setresuid(pw->pw_uid, pw->pw_uid, pw->pw_uid) == -1)
+		fatal("failed to drop privileges");
+
 #ifdef PROFILE
 	if (unveil("gmon.out", "rwc") != 0)
 		err(1, "gmon.out");
 #endif
 
-	if (unveil(env->httpd_chroot, "r") == -1)
-		err(1, "unveil");
-
 	if (unveil(GOTWEBD_CONF, "r") == -1)
 		err(1, "unveil");
 
@@ -420,6 +424,7 @@ main(int argc, char **argv)
 #endif
 
 	event_dispatch();
+	event_base_free(evb);
 
 	log_debug("%s gotwebd exiting", getprogname());
 
@@ -480,10 +485,11 @@ gotwebd_shutdown(void)
 
 	for (i = 0; i < env->nserver; ++i) {
 		event_del(&env->iev_server[i].ev);
-		imsg_clear(&env->iev_server[i].ibuf);
+		imsgbuf_clear(&env->iev_server[i].ibuf);
 		close(env->iev_server[i].ibuf.fd);
 		env->iev_server[i].ibuf.fd = -1;
 	}
+	free(env->iev_server);
 
 	do {
 		pid = waitpid(WAIT_ANY, &status, 0);
@@ -498,6 +504,21 @@ gotwebd_shutdown(void)
 			    pid);
 	} while (pid != -1 || (pid == -1 && errno == EINTR));
 
+	while (!TAILQ_EMPTY(&gotwebd_env->addresses)) {
+		struct address *h;
+
+		h = TAILQ_FIRST(&gotwebd_env->addresses);
+		TAILQ_REMOVE(&gotwebd_env->addresses, h, entry);
+		free(h);
+	}
+	while (!TAILQ_EMPTY(&gotwebd_env->servers)) {
+		struct server *srv;
+
+		srv = TAILQ_FIRST(&gotwebd_env->servers);
+		TAILQ_REMOVE(&gotwebd_env->servers, srv, entry);
+		free(srv);
+	}
+	sockets_purge(gotwebd_env);
 	free(gotwebd_env);
 
 	log_warnx("gotwebd terminating");
