@@ -63,9 +63,6 @@
 
 #ifndef __GNUC__
 #define __attribute__(a)
-#ifndef __bounded__
-#define __bounded__(a, b, c)
-#endif
 #endif
 
 
@@ -110,6 +107,10 @@
 #define __unused __attribute__ ((__unused__))
 #endif
 
+#if !defined(__bounded__) && !defined(__OpenBSD__)
+#define __bounded__(a, b, c)
+#endif
+
 #ifndef __OpenBSD__
 #define pledge(s, p) (0)
 #define unveil(s, p) (0)
