@@ -91,6 +91,7 @@ static const struct got_error got_errors[] = {
 	{ GOT_ERR_PACK_OFFSET,	"bad offset in pack file" },
 	{ GOT_ERR_OBJ_EXISTS,	"object already exists" },
 	{ GOT_ERR_BAD_OBJ_ID,	"bad object id" },
+	{ GOT_ERR_OBJECT_FORMAT, "object format not supported" },
 	{ GOT_ERR_ITER_COMPLETED,"iteration completed" },
 	{ GOT_ERR_RANGE,	"value out of range" },
 	{ GOT_ERR_EXPECTED,	"expected an error but have no error" },
@@ -244,7 +245,10 @@ static const struct got_error got_errors[] = {
 	    "the work tree is no longer up-to-date; merge must be aborted "
 	    "and retried" },
 	{ GOT_ERR_BUNDLE_FORMAT, "unknown git bundle version" },
-	{ GOT_ERR_BAD_KEYWORD, "invalid commit keyword" }
+	{ GOT_ERR_BAD_KEYWORD, "invalid commit keyword" },
+	{ GOT_ERR_UNKNOWN_CAPA, "unknown capability" },
+	{ GOT_ERR_REF_DUP_ENTRY, "duplicate reference entry" },
+	{ GOT_ERR_DIFF_NOCHANGES, "no changes match the requested diff" },
 };
 
 static struct got_custom_error {
