@@ -16,9 +16,9 @@
  */
 
 /*
- * All code runs under the same UID but sensitive code paths are
- * run in a separate process with tighter pledge(2) promises.
- * Data is communicated between processes via imsg_flush(3) and imsg_read(3).
+ * All code runs under the same UID but sensitive code paths are run in a
+ * separate process with tighter pledge(2) promises; data is communicated
+ * between processes via imsgbuf_flush(3) and imsgbuf_read(3).
  * This behaviour is transparent to users of the library.
  *
  * We generally transmit data in imsg buffers, split across several messages
@@ -82,7 +82,7 @@
 	GOT_STRINGVAL(GOT_LIBEXECDIR) "/" GOT_STRINGVAL(GOT_PROG_FETCH_HTTP)
 
 enum got_imsg_type {
-	/* An error occured while processing a request. */
+	/* An error occurred while processing a request. */
 	GOT_IMSG_ERROR,
 
 	/* Stop the child process. */
@@ -190,7 +190,7 @@ enum got_imsg_type {
 	GOT_IMSG_RAW_DELTA_REQUEST,
 	GOT_IMSG_RAW_DELTA,
 
-	/* Re-use deltas found in a pack file. */
+	/* Reuse deltas found in a pack file. */
 	GOT_IMSG_DELTA_REUSE_REQUEST,
 	GOT_IMSG_REUSED_DELTAS,
 	GOT_IMSG_DELTA_REUSE_DONE,
@@ -260,7 +260,8 @@ struct got_imsg_commit_object {
 } __attribute__((__packed__));
 
 struct got_imsg_tree_entry {
-	char id[SHA1_DIGEST_LENGTH];
+	char id[GOT_HASH_DIGEST_MAXLEN];
+	int algo;
 	mode_t mode;
 	size_t namelen;
 	/* Followed by namelen bytes of entry's name, not NUL-terminated. */
@@ -345,13 +346,6 @@ struct got_imsg_reused_deltas {
 	/ sizeof(struct got_imsg_reused_delta))
 };
 
-/* Structure for GOT_IMSG_COMMIT_PAINTING_REQUEST. */
-struct got_imsg_commit_painting_request {
-	struct got_object_id id;
-	int idx;
-	int color;
-} __attribute__((__packed__));
-
 /* Structure for GOT_IMSG_PAINTED_COMMITS. */
 struct got_imsg_painted_commit {
 	struct got_object_id id;
@@ -491,7 +485,7 @@ struct got_imsg_send_ref_status {
 
 /* Structure for GOT_IMSG_IDXPACK_REQUEST data. */
 struct got_imsg_index_pack_request {
-	uint8_t pack_hash[SHA1_DIGEST_LENGTH];
+	struct got_object_id id;
 } __attribute__((__packed__));
 
 /* Structure for GOT_IMSG_IDXPACK_PROGRESS data. */
@@ -513,14 +507,16 @@ struct got_imsg_index_pack_progress {
 struct got_imsg_packidx {
 	size_t len;
 	off_t packfile_size;
-	/* Additionally, a file desciptor is passed via imsg. */
+	int algo;
+	/* Additionally, a file descriptor is passed via imsg. */
 };
 
 /* Structure for GOT_IMSG_PACK. */
 struct got_imsg_pack {
 	char path_packfile[PATH_MAX];
 	off_t filesize;
-	/* Additionally, a file desciptor is passed via imsg. */
+	int algo;
+	/* Additionally, a file descriptor is passed via imsg. */
 } __attribute__((__packed__));
 
 /*
@@ -652,8 +648,8 @@ struct got_imsg_patch {
 	int	xbit;
 	char	old[PATH_MAX];
 	char	new[PATH_MAX];
-	char	cid[41];
-	char	blob[41];
+	char	cid[GOT_HASH_DIGEST_STRING_MAXLEN];
+	char	blob[GOT_HASH_DIGEST_STRING_MAXLEN];
 };
 
 /*
@@ -698,7 +694,7 @@ const struct got_error *got_privsep_send_tmpfd(struct imsgbuf *, int);
 const struct got_error *got_privsep_send_obj(struct imsgbuf *,
     struct got_object *);
 const struct got_error *got_privsep_send_index_pack_req(struct imsgbuf *,
-    uint8_t *, int);
+    struct got_object_id *, int);
 const struct got_error *got_privsep_send_index_pack_outfd(struct imsgbuf *,
     int);
 const struct got_error *got_privsep_recv_index_progress(int *, int *, int *,
@@ -830,8 +826,7 @@ const struct got_error *got_privsep_recv_reused_deltas(int *,
     struct got_imsg_reused_delta *, size_t *, struct imsgbuf *);
 
 const struct got_error *got_privsep_init_commit_painting(struct imsgbuf *);
-const struct got_error *got_privsep_send_painting_request(struct imsgbuf *,
-    int, struct got_object_id *, intptr_t);
+const struct got_error *got_privsep_send_painting_request(struct imsgbuf *);
 typedef const struct got_error *(*got_privsep_recv_painted_commit_cb)(void *,
     struct got_object_id *, intptr_t);
 const struct got_error *got_privsep_send_painted_commits(struct imsgbuf *,
