@@ -297,8 +297,10 @@ parse_packed_ref_line(struct got_reference **ref, const char *abs_refname,
 {
 	struct got_object_id id;
 	const char *name;
+	size_t digest_string_len;
 
 	*ref = NULL;
+	digest_string_len = got_hash_digest_string_length(algo);
 
 	if (line[0] == '#' || line[0] == '^')
 		return NULL;
@@ -307,11 +309,11 @@ parse_packed_ref_line(struct got_reference **ref, const char *abs_refname,
 		return got_error(GOT_ERR_BAD_REF_DATA);
 
 	if (abs_refname) {
-		if (strcmp(line + SHA1_DIGEST_STRING_LENGTH, abs_refname) != 0)
+		if (strcmp(line + digest_string_len, abs_refname) != 0)
 			return NULL;
 		name = abs_refname;
 	} else
-		name = line + SHA1_DIGEST_STRING_LENGTH;
+		name = line + digest_string_len;
 
 	return alloc_ref(ref, name, &id, GOT_REF_IS_PACKED, mtime);
 }
@@ -641,6 +643,12 @@ got_ref_get_mtime(struct got_reference *ref)
 	return ref->mtime;
 }
 
+time_t
+got_ref_get_cached_committer_time(struct got_reference *ref)
+{
+	return ref->committer_time;
+}
+
 const struct got_error *
 got_ref_cmp_by_name(void *arg, int *cmp, struct got_reference *re1,
     struct got_reference* re2)
@@ -1308,7 +1316,7 @@ delete_packed_ref(struct got_reference *delref, struct got_repository *repo)
 	struct got_reflist_head refs;
 	int found_delref = 0;
 
-	/* The packed-refs file does not cotain symbolic references. */
+	/* The packed-refs file does not contain symbolic references. */
 	if (delref->flags & GOT_REF_IS_SYMBOLIC)
 		return got_error(GOT_ERR_BAD_REF_DATA);
 
