@@ -51,13 +51,13 @@
 
 #include "got_lib_delta.h"
 #include "got_lib_delta_cache.h"
+#include "got_lib_hash.h"
 #include "got_lib_inflate.h"
 #include "got_lib_object.h"
 #include "got_lib_object_parse.h"
 #include "got_lib_object_create.h"
 #include "got_lib_pack.h"
 #include "got_lib_privsep.h"
-#include "got_lib_hash.h"
 #include "got_lib_object_cache.h"
 #include "got_lib_repository.h"
 #include "got_lib_gotconfig.h"
@@ -708,7 +708,7 @@ got_repo_open(struct got_repository **repop, const char *path,
 		return got_error_from_errno("calloc");
 
 	RB_INIT(&repo->packidx_bloom_filters);
-	TAILQ_INIT(&repo->packidx_paths);
+	RB_INIT(&repo->packidx_paths);
 
 	for (i = 0; i < nitems(repo->privsep_children); i++) {
 		memset(&repo->privsep_children[i], 0,
@@ -787,7 +787,7 @@ got_repo_open(struct got_repository **repop, const char *path,
 	err = read_gitconfig(repo, global_gitconfig_path);
 	if (err)
 		goto done;
-	if (repo->gitconfig_repository_format_version != 0) {
+	if (repo->gitconfig_repository_format_version > 1) {
 		err = got_error_path(path, GOT_ERR_GIT_REPO_FORMAT);
 		goto done;
 	}
@@ -796,6 +796,18 @@ got_repo_open(struct got_repository **repop, const char *path,
 		char *val = repo->extvals[i];
 		int j, supported = 0;
 
+		if (repo->gitconfig_repository_format_version == 1 &&
+		    strcasecmp(ext, "objectformat") == 0) {
+			if (strcmp(val, "sha1") == 0)
+				continue;
+			if (strcmp(val, "sha256") == 0) {
+				repo->algo = GOT_HASH_SHA256;
+				continue;
+			}
+			err = got_error_path(val, GOT_ERR_OBJECT_FORMAT);
+			goto done;
+		}
+
 		if (!is_boolean_val(val)) {
 			err = got_error_path(ext, GOT_ERR_GIT_REPO_EXT);
 			goto done;
@@ -865,7 +877,7 @@ got_repo_close(struct got_repository *repo)
 	for (i = 0; i < nitems(repo->privsep_children); i++) {
 		if (repo->privsep_children[i].imsg_fd == -1)
 			continue;
-		imsg_clear(repo->privsep_children[i].ibuf);
+		imsgbuf_clear(repo->privsep_children[i].ibuf);
 		free(repo->privsep_children[i].ibuf);
 		err = got_privsep_send_stop(repo->privsep_children[i].imsg_fd);
 		if (err && err->code == GOT_ERR_EOF)
@@ -914,7 +926,7 @@ got_repo_remote_repo_dup(struct got_remote_repo **newp,
 	new = calloc(1, sizeof(*new));
 	if (new == NULL)
 		return got_error_from_errno("calloc");
-	
+
 	if (repo->name) {
 		new->name = strdup(repo->name);
 		if (new->name == NULL) {
@@ -1178,16 +1190,21 @@ cache_packidx(struct got_repository *repo, struct got_packidx *packidx,
 }
 
 int
-got_repo_is_packidx_filename(const char *name, size_t len)
+got_repo_is_packidx_filename(const char *name, size_t len,
+    enum got_hash_algorithm algo)
 {
-	if (len != GOT_PACKIDX_NAMELEN)
+	size_t digest_string_len;
+
+	digest_string_len = got_hash_digest_string_length(algo);
+
+	if (len != GOT_PACKIDX_NAMELEN(digest_string_len))
 		return 0;
 
 	if (strncmp(name, GOT_PACK_PREFIX, strlen(GOT_PACK_PREFIX)) != 0)
 		return 0;
 
-	if (strcmp(name + strlen(GOT_PACK_PREFIX) +
-	    SHA1_DIGEST_STRING_LENGTH - 1, GOT_PACKIDX_SUFFIX) != 0)
+	if (strcmp(name + strlen(GOT_PACK_PREFIX) + digest_string_len - 1,
+	    GOT_PACKIDX_SUFFIX) != 0)
 		return 0;
 
 	return 1;
@@ -1215,7 +1232,8 @@ got_repo_check_packidx_bloom_filter(struct got_repository *repo,
 
 	bf = get_packidx_bloom_filter(repo, path_packidx, strlen(path_packidx));
 	if (bf)
-		return bloom_check(bf->bloom, id->sha1, sizeof(id->sha1));
+		return bloom_check(bf->bloom, id->hash,
+		    got_hash_digest_length(id->algo));
 
 	/* No bloom filter means this pack index must be searched. */
 	return 1;
@@ -1227,7 +1245,9 @@ add_packidx_bloom_filter(struct got_repository *repo,
 {
 	int i, nobjects = be32toh(packidx->hdr.fanout_table[0xff]);
 	struct got_packidx_bloom_filter *bf;
-	size_t len;
+	size_t len, digest_len;
+
+	digest_len = got_hash_digest_length(repo->algo);
 
 	/*
 	 * Don't use bloom filters for very large pack index files.
@@ -1265,9 +1285,8 @@ add_packidx_bloom_filter(struct got_repository *repo,
 	/* Minimum size supported by our bloom filter is 1000 entries. */
 	bloom_init(bf->bloom, nobjects < 1000 ? 1000 : nobjects, 0.1);
 	for (i = 0; i < nobjects; i++) {
-		struct got_packidx_object_id *id;
-		id = &packidx->hdr.sorted_ids[i];
-		bloom_add(bf->bloom, id->sha1, sizeof(id->sha1));
+		uint8_t *id = packidx->hdr.sorted_ids + i * digest_len;
+		bloom_add(bf->bloom, id, digest_len);
 	}
 
 	RB_INSERT(got_packidx_bloom_filter_tree,
@@ -1280,9 +1299,9 @@ purge_packidx_paths(struct got_pathlist_head *packidx_paths)
 {
 	struct got_pathlist_entry *pe;
 
-	while (!TAILQ_EMPTY(packidx_paths)) {
-		pe = TAILQ_FIRST(packidx_paths);
-		TAILQ_REMOVE(packidx_paths, pe, entry);
+	while (!RB_EMPTY(packidx_paths)) {
+		pe = RB_MIN(got_pathlist_head, packidx_paths);
+		RB_REMOVE(got_pathlist_head, packidx_paths, pe);
 		free((char *)pe->path);
 		free(pe);
 	}
@@ -1304,7 +1323,7 @@ refresh_packidx_paths(struct got_repository *repo)
 			err = got_error_from_errno2("stat", objects_pack_dir);
 			goto done;
 		}
-	} else if (TAILQ_EMPTY(&repo->packidx_paths) ||
+	} else if (RB_EMPTY(&repo->packidx_paths) ||
 	    sb.st_mtim.tv_sec != repo->pack_path_mtime.tv_sec ||
 	    sb.st_mtim.tv_nsec != repo->pack_path_mtime.tv_nsec) {
 		purge_packidx_paths(&repo->packidx_paths);
@@ -1359,7 +1378,7 @@ got_repo_search_packidx(struct got_packidx **packidx, int *idx,
 	if (err)
 		return err;
 
-	TAILQ_FOREACH(pe, &repo->packidx_paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, &repo->packidx_paths) {
 		const char *path_packidx = pe->path;
 		int is_cached = 0;
 
@@ -1380,9 +1399,14 @@ got_repo_search_packidx(struct got_packidx **packidx, int *idx,
 			continue; /* already searched */
 
 		err = got_packidx_open(packidx, got_repo_get_fd(repo),
-		    path_packidx, 0);
-		if (err)
+		    path_packidx, 0, repo->algo);
+		if (err) {
+			if (err->code == GOT_ERR_LONELY_PACKIDX) {
+				err = NULL;
+				continue;
+			}
 			goto done;
+		}
 
 		err = add_packidx_bloom_filter(repo, *packidx, path_packidx);
 		if (err)
@@ -1439,7 +1463,8 @@ got_repo_list_packidx(struct got_pathlist_head *packidx_paths,
 
 	while ((dent = readdir(packdir)) != NULL) {
 		if (!got_repo_is_packidx_filename(dent->d_name,
-		    strlen(dent->d_name)))
+		    strlen(dent->d_name),
+		    repo->algo))
 			continue;
 
 		if (asprintf(&path_packidx, "%s/%s", GOT_OBJECTS_PACK_DIR,
@@ -1449,7 +1474,7 @@ got_repo_list_packidx(struct got_pathlist_head *packidx_paths,
 			break;
 		}
 
-		err = got_pathlist_append(packidx_paths, path_packidx, NULL);
+		err = got_pathlist_insert(NULL, packidx_paths, path_packidx, NULL);
 		if (err)
 			break;
 	}
@@ -1483,7 +1508,7 @@ got_repo_get_packidx(struct got_packidx **packidx, const char *path_packidx,
 	/* No luck. Search the filesystem. */
 
 	err = got_packidx_open(packidx, got_repo_get_fd(repo),
-	    path_packidx, 0);
+	    path_packidx, 0, repo->algo);
 	if (err)
 		return err;
 
@@ -1551,6 +1576,7 @@ got_repo_cache_pack(struct got_pack **packp, struct got_repository *repo,
 {
 	const struct got_error *err = NULL;
 	struct got_pack *pack = NULL;
+	struct got_pack tmp;
 	struct stat sb;
 	size_t i;
 
@@ -1566,7 +1592,6 @@ got_repo_cache_pack(struct got_pack **packp, struct got_repository *repo,
 	}
 
 	if (i == repo->pack_cache_size) {
-		struct got_pack tmp;
 		do {
 			i--;
 		} while (i > 0 && repo->pinned_pack >= 0 &&
@@ -1578,6 +1603,9 @@ got_repo_cache_pack(struct got_pack **packp, struct got_repository *repo,
 			return got_error_from_errno("ftruncate");
 		if (ftruncate(repo->packs[i].accumfd, 0L) == -1)
 			return got_error_from_errno("ftruncate");
+	}
+
+	if (i != 0) {
 		memcpy(&tmp, &repo->packs[i], sizeof(tmp));
 		memcpy(&repo->packs[i], &repo->packs[0],
 		    sizeof(repo->packs[i]));
@@ -1606,6 +1634,7 @@ got_repo_cache_pack(struct got_pack **packp, struct got_repository *repo,
 		goto done;
 	}
 	pack->filesize = sb.st_size;
+	pack->algo = repo->algo;
 
 	pack->privsep_child = NULL;
 
@@ -1699,7 +1728,8 @@ got_repo_unpin_pack(struct got_repository *repo)
 }
 
 const struct got_error *
-got_repo_init(const char *repo_path, const char *head_name)
+got_repo_init(const char *repo_path, const char *head_name,
+    enum got_hash_algorithm algo)
 {
 	const struct got_error *err = NULL;
 	const char *dirnames[] = {
@@ -1710,13 +1740,23 @@ got_repo_init(const char *repo_path, const char *head_name)
 	const char *description_str = "Unnamed repository; "
 	    "edit this file 'description' to name the repository.";
 	const char *headref = "ref: refs/heads/";
-	const char *gitconfig_str = "[core]\n"
+	const char *gitconfig_sha1 = "[core]\n"
 	    "\trepositoryformatversion = 0\n"
 	    "\tfilemode = true\n"
 	    "\tbare = true\n";
+	const char *gitconfig_sha256 = "[core]\n"
+	    "\trepositoryformatversion = 1\n"
+	    "\tfilemode = true\n"
+	    "\tbare = true\n"
+	    "[extensions]\n"
+	    "\tobjectformat = sha256\n";
+	const char *gitconfig = gitconfig_sha1;
 	char *headref_str, *path;
 	size_t i;
 
+	if (algo == GOT_HASH_SHA256)
+		gitconfig = gitconfig_sha256;
+
 	if (!got_path_dir_is_empty(repo_path))
 		return got_error(GOT_ERR_DIR_NOT_EMPTY);
 
@@ -1752,7 +1792,7 @@ got_repo_init(const char *repo_path, const char *head_name)
 
 	if (asprintf(&path, "%s/%s", repo_path, "config") == -1)
 		return got_error_from_errno("asprintf");
-	err = got_path_create_file(path, gitconfig_str);
+	err = got_path_create_file(path, gitconfig);
 	free(path);
 	if (err)
 		return err;
@@ -1790,7 +1830,7 @@ retry:
 	tv.tv_sec = repo->pack_path_mtime.tv_sec;
 	tv.tv_nsec = repo->pack_path_mtime.tv_nsec;
 
-	TAILQ_FOREACH(pe, &repo->packidx_paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head, &repo->packidx_paths) {
 		const char *path_packidx;
 		struct got_packidx *packidx;
 		struct got_object_qid *qid;
@@ -1813,9 +1853,14 @@ retry:
 		path_packidx = pe->path;
 
 		err = got_packidx_open(&packidx, got_repo_get_fd(repo),
-		    path_packidx, 0);
-		if (err)
+		    path_packidx, 0, repo->algo);
+		if (err) {
+			if (err->code == GOT_ERR_LONELY_PACKIDX) {
+				err = NULL;
+				continue;
+			}
 			break;
+		}
 
 		got_object_id_queue_free(&matched_ids);
 
@@ -1954,15 +1999,16 @@ got_repo_match_object_id_prefix(struct got_object_id **id,
 {
 	const struct got_error *err = NULL;
 	char *path_objects = NULL, *object_dir = NULL;
-	size_t len;
+	size_t len, digest_string_len;
 	int i;
 
 	*id = NULL;
 
 	path_objects = got_repo_get_path_objects(repo);
+	digest_string_len = got_hash_digest_string_length(repo->algo);
 
 	len = strlen(id_str_prefix);
-	if (len > SHA1_DIGEST_STRING_LENGTH - 1) {
+	if (len > digest_string_len - 1) {
 		err = got_error_path(id_str_prefix, GOT_ERR_BAD_OBJ_ID_STR);
 		goto done;
 	}
@@ -2165,6 +2211,7 @@ got_repo_find_object_id(struct got_object_id *id, struct got_repository *repo)
 
 	err = got_repo_match_object_id_prefix(&matched_id, id_str,
 	    GOT_OBJ_TYPE_ANY, repo);
+	free(matched_id);
 	free(id_str);
 	return err;
 }
@@ -2233,6 +2280,41 @@ done:
 	return err;
 }
 
+static const struct got_error *
+insert_dir_entry(struct dirent *de, long pos,
+    struct got_pathlist_head *dirents)
+{
+	const struct got_error *err = NULL;
+	struct got_pathlist_entry *new_pe;
+	char *d_name;
+	long *ppos;
+
+	d_name = strdup(de->d_name);
+	if (d_name == NULL)
+		return got_error_from_errno("strdup");
+
+	/* Record the offset to this dirent in struct DIR as path data. */
+	ppos = malloc(sizeof(*ppos));
+	if (ppos == NULL) {
+		err = got_error_from_errno("malloc");
+		goto done;
+	}
+	*ppos = pos;
+
+	err = got_pathlist_insert(&new_pe, dirents, d_name, ppos);
+	if (err)
+		goto done;
+
+	if (new_pe == NULL)
+		err = got_error(GOT_ERR_TREE_DUP_ENTRY);
+done:
+	if (err) {
+		free(d_name);
+		free(ppos);
+	}
+	return err;
+}
+
 static const struct got_error *
 insert_tree_entry(struct got_tree_entry *new_te,
     struct got_pathlist_head *paths)
@@ -2291,6 +2373,37 @@ done:
 	return err;
 }
 
+static const struct got_error *
+match_ignores(int *ignore, char *d_name, int type,
+    struct got_pathlist_head *ignores)
+{
+	struct got_pathlist_entry *pe;
+
+	*ignore = 0;
+
+	RB_FOREACH(pe, got_pathlist_head, ignores) {
+		if (type == DT_DIR && pe->path_len > 0 &&
+		    pe->path[pe->path_len - 1] == '/') {
+			char stripped[PATH_MAX];
+
+			if (strlcpy(stripped, pe->path,
+			    sizeof(stripped)) >= sizeof(stripped))
+				return got_error(GOT_ERR_NO_SPACE);
+
+			got_path_strip_trailing_slashes(stripped);
+			if (fnmatch(stripped, d_name, 0) == 0) {
+				*ignore = 1;
+				break;
+			}
+		} else if (fnmatch(pe->path, d_name, 0) == 0) {
+			*ignore = 1;
+			break;
+		}
+	}
+
+	return NULL;
+}
+
 static const struct got_error *
 write_tree(struct got_object_id **new_tree_id, const char *path_dir,
     struct got_pathlist_head *ignores, struct got_repository *repo,
@@ -2301,12 +2414,14 @@ write_tree(struct got_object_id **new_tree_id, const char *path_dir,
 	struct dirent *de;
 	int nentries;
 	struct got_tree_entry *new_te = NULL;
-	struct got_pathlist_head paths;
+	struct got_pathlist_head paths, dirents;
 	struct got_pathlist_entry *pe;
+	long pos;
 
 	*new_tree_id = NULL;
 
-	TAILQ_INIT(&paths);
+	RB_INIT(&paths);
+	RB_INIT(&dirents);
 
 	dir = opendir(path_dir);
 	if (dir == NULL) {
@@ -2314,39 +2429,53 @@ write_tree(struct got_object_id **new_tree_id, const char *path_dir,
 		goto done;
 	}
 
-	nentries = 0;
+	/*
+	 * Sort dirents into a pathlist. Otherwise subdir-iteration order
+	 * depends on the readdir() function, which varies between different
+	 * operating systems, making regression testing more difficult.
+	 * This smells of TOCTOU, but if dirents appear, disappear, or change
+	 * during import then we have a racy situation no matter what.
+	 */
+	pos = telldir(dir);
+	if (pos == -1) {
+		err = got_error_from_errno2("telldir", path_dir);
+		goto done;
+	}
 	while ((de = readdir(dir)) != NULL) {
+		if (strcmp(de->d_name, ".") != 0 &&
+		    strcmp(de->d_name, "..") != 0) {
+			err = insert_dir_entry(de, pos, &dirents);
+			if (err)
+				goto done;
+		}
+		pos = telldir(dir);
+		if (pos == -1) {
+			err = got_error_from_errno2("telldir", path_dir);
+			goto done;
+		}
+	}
+
+	nentries = 0;
+	RB_FOREACH(pe, got_pathlist_head, &dirents) {
 		int ignore = 0;
 		int type;
+		long *pos = pe->data;
 
-		if (strcmp(de->d_name, ".") == 0 ||
-		    strcmp(de->d_name, "..") == 0)
-			continue;
+		seekdir(dir, *pos);
+		de = readdir(dir);
+		if (de == NULL) { /* should not happen */
+			err = got_error_fmt(GOT_ERR_EOF,
+			    "unexpected EOF on directory '%s' while seeking "
+			    "to entry '%s' at offset '%ld'\n", path_dir,
+			    pe->path, *pos);
+			goto done;
+		}
 
 		err = got_path_dirent_type(&type, path_dir, de);
 		if (err)
 			goto done;
 
-		TAILQ_FOREACH(pe, ignores, entry) {
-			if (type == DT_DIR && pe->path_len > 0 &&
-			    pe->path[pe->path_len - 1] == '/') {
-				char stripped[PATH_MAX];
-
-				if (strlcpy(stripped, pe->path,
-				    sizeof(stripped)) >= sizeof(stripped)) {
-					err = got_error(GOT_ERR_NO_SPACE);
-					goto done;
-				}
-				got_path_strip_trailing_slashes(stripped);
-				if (fnmatch(stripped, de->d_name, 0) == 0) {
-					ignore = 1;
-					break;
-				}
-			} else if (fnmatch(pe->path, de->d_name, 0) == 0) {
-				ignore = 1;
-				break;
-			}
-		}
+		err = match_ignores(&ignore, de->d_name, type, ignores);
 		if (ignore)
 			continue;
 
@@ -2372,13 +2501,13 @@ write_tree(struct got_object_id **new_tree_id, const char *path_dir,
 		nentries++;
 	}
 
-	if (TAILQ_EMPTY(&paths)) {
+	if (RB_EMPTY(&paths)) {
 		err = got_error_msg(GOT_ERR_NO_TREE_ENTRY,
 		    "cannot create tree without any entries");
 		goto done;
 	}
 
-	TAILQ_FOREACH(pe, &paths, entry) {
+	RB_FOREACH(pe, got_pathlist_head,  &paths) {
 		struct got_tree_entry *te = pe->data;
 		char *path;
 		if (!S_ISREG(te->mode) && !S_ISLNK(te->mode))
@@ -2397,6 +2526,7 @@ write_tree(struct got_object_id **new_tree_id, const char *path_dir,
 done:
 	if (dir)
 		closedir(dir);
+	got_pathlist_free(&dirents, GOT_PATHLIST_FREE_ALL);
 	got_pathlist_free(&paths, GOT_PATHLIST_FREE_NONE);
 	return err;
 }
@@ -2550,7 +2680,8 @@ got_repo_get_packfile_info(int *npackfiles, int *nobjects,
 
 	while ((dent = readdir(packdir)) != NULL) {
 		if (!got_repo_is_packidx_filename(dent->d_name,
-		    strlen(dent->d_name)))
+		    strlen(dent->d_name),
+		    repo->algo))
 			continue;
 
 		if (asprintf(&path_packidx, "%s/%s", GOT_OBJECTS_PACK_DIR,
@@ -2560,10 +2691,15 @@ got_repo_get_packfile_info(int *npackfiles, int *nobjects,
 		}
 
 		err = got_packidx_open(&packidx, got_repo_get_fd(repo),
-		    path_packidx, 0);
+		    path_packidx, 0, repo->algo);
 		free(path_packidx);
-		if (err)
+		if (err) {
+			if (err->code == GOT_ERR_LONELY_PACKIDX) {
+				err = NULL;
+				continue;
+			}
 			goto done;
+		}
 
 		if (fstat(packidx->fd, &sb) == -1)
 			goto done;
