@@ -50,6 +50,12 @@
 #define nitems(_a)	(sizeof((_a)) / sizeof((_a)[0]))
 #endif
 
+/*
+ * Number of seconds until we give up trying to read more
+ * data from the client connection.
+ */
+static const int timeout = 60;
+
 static const struct got_capability read_capabilities[] = {
 	{ GOT_CAPA_AGENT, "got/" GOT_VERSION_STR },
 	{ GOT_CAPA_OFS_DELTA, NULL },
@@ -408,7 +414,6 @@ send_capability(struct got_capability *capa, struct imsgbuf *ibuf)
 			return got_error_from_errno("imsg_add CAPABILITY");
 	}
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 
 	return NULL;
@@ -666,6 +671,7 @@ recv_done(int *packfd, int outfd, struct imsgbuf *ibuf, int chattygot)
 {
 	const struct got_error *err;
 	struct imsg imsg;
+	int fd;
 
 	*packfd = -1;
 
@@ -686,10 +692,17 @@ recv_done(int *packfd, int outfd, struct imsgbuf *ibuf, int chattygot)
 			err = gotd_imsg_recv_error(NULL, &imsg);
 			break;
 		case GOTD_IMSG_PACKFILE_PIPE:
-			if (imsg.fd != -1)
-				*packfd = imsg.fd;
-			else
+			fd = imsg_get_fd(&imsg);
+			if (fd == -1) {
 				err = got_error(GOT_ERR_PRIVSEP_NO_FD);
+				break;
+			}
+			*packfd = fd;
+			if (imsg_compose(ibuf, GOTD_IMSG_PACKFILE_READY,
+			    0, 0, -1, NULL, 0) == -1)
+				err = got_error_from_errno("imsg_compose "
+				    "PACKFILE_READY");
+			err = gotd_imsg_flush(ibuf);
 			break;
 		default:
 			err = got_error(GOT_ERR_PRIVSEP_MSG);
@@ -773,8 +786,7 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 	enum protostate {
 		STATE_EXPECT_WANT,
 		STATE_EXPECT_MORE_WANT,
-		STATE_EXPECT_HAVE,
-		STATE_EXPECT_DONE,
+		STATE_EXPECT_HAVE_OR_DONE,
 		STATE_DONE,
 	};
 	enum protostate curstate = STATE_EXPECT_WANT;
@@ -782,7 +794,9 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 	int packfd = -1;
 	size_t pack_chunksize;
 
-	imsg_init(&ibuf, gotd_sock);
+	if (imsgbuf_init(&ibuf, gotd_sock) == -1)
+		return got_error_from_errno("imsgbuf_init");
+	imsgbuf_allow_fdpass(&ibuf);
 
 	err = announce_refs(outfd, &ibuf, 1, repo_path, chattygot);
 	if (err)
@@ -791,14 +805,14 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 	while (curstate != STATE_DONE) {
 		int n;
 		buf[0] = '\0';
-		err = got_pkt_readpkt(&n, infd, buf, sizeof(buf), chattygot);
+		err = got_pkt_readpkt(&n, infd, buf, sizeof(buf), chattygot,
+		    timeout);
 		if (err)
 			goto done;
 		if (n == 0) {
 			if (curstate != STATE_EXPECT_WANT &&
 			    curstate != STATE_EXPECT_MORE_WANT &&
-			    curstate != STATE_EXPECT_HAVE &&
-			    curstate != STATE_EXPECT_DONE) {
+			    curstate != STATE_EXPECT_HAVE_OR_DONE) {
 				err = got_error_msg(GOT_ERR_BAD_PACKET,
 				    "unexpected flush packet received");
 				goto done;
@@ -827,20 +841,19 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 
 			if (curstate == STATE_EXPECT_WANT ||
 			    curstate == STATE_EXPECT_MORE_WANT ||
-			    curstate == STATE_EXPECT_HAVE) {
+			    curstate == STATE_EXPECT_HAVE_OR_DONE) {
 				err = forward_flushpkt(&ibuf);
 				if (err)
 					goto done;
 			}
-			if (curstate == STATE_EXPECT_HAVE && !have_ack) {
+			if (curstate == STATE_EXPECT_HAVE_OR_DONE &&
+			    !have_ack) {
 				err = send_nak(outfd, chattygot);
 				if (err)
 					goto done;
 			}
 			if (curstate == STATE_EXPECT_MORE_WANT)
-				curstate = STATE_EXPECT_HAVE;
-			else
-				curstate = STATE_EXPECT_DONE;
+				curstate = STATE_EXPECT_HAVE_OR_DONE;
 		} else if (n >= 5 && strncmp(buf, "want ", 5) == 0) {
 			if (curstate != STATE_EXPECT_WANT &&
 			    curstate != STATE_EXPECT_MORE_WANT) {
@@ -855,22 +868,18 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 			if (curstate == STATE_EXPECT_WANT)
 				curstate = STATE_EXPECT_MORE_WANT;
 		} else if (n >= 5 && strncmp(buf, "have ", 5) == 0) {
-			if (curstate != STATE_EXPECT_HAVE &&
-			    curstate != STATE_EXPECT_DONE) {
+			if (curstate != STATE_EXPECT_HAVE_OR_DONE) {
 				err = got_error_msg(GOT_ERR_BAD_PACKET,
 				    "unexpected 'have' packet");
 				goto done;
 			}
-			if (curstate == STATE_EXPECT_HAVE) {
-				err = recv_have(&have_ack, outfd, &ibuf,
-				    buf, n, chattygot);
-				if (err)
-					goto done;
-				seen_have = 1;
-			}
+			err = recv_have(&have_ack, outfd, &ibuf,
+			    buf, n, chattygot);
+			if (err)
+				goto done;
+			seen_have = 1;
 		} else if (n == 5 && strncmp(buf, "done\n", 5) == 0) {
-			if (curstate != STATE_EXPECT_HAVE &&
-			    curstate != STATE_EXPECT_DONE) {
+			if (curstate != STATE_EXPECT_HAVE_OR_DONE) {
 				err = got_error_msg(GOT_ERR_BAD_PACKET,
 				    "unexpected 'done' packet");
 				goto done;
@@ -928,7 +937,7 @@ serve_read(int infd, int outfd, int gotd_sock, const char *repo_path,
 		}
 	}
 done:
-	imsg_clear(&ibuf);
+	imsgbuf_clear(&ibuf);
 	if (packfd != -1 && close(packfd) == -1 && err == NULL)
 		err = got_error_from_errno("close");
 	if (err)
@@ -1024,7 +1033,6 @@ recv_ref_update(int *report_status, int outfd, struct imsgbuf *ibuf,
 		return got_error_from_errno("imsg_add REF_UPDATE");
 	if (imsg_add(wbuf, refname, iref.name_len) == -1)
 		return got_error_from_errno("imsg_add REF_UPDATE");
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 
 	err = gotd_imsg_flush(ibuf);
@@ -1072,10 +1080,10 @@ recv_packfile(struct imsg *imsg, int infd)
 	if (datalen != 0)
 		return got_error(GOT_ERR_PRIVSEP_MSG);
 
-	if (imsg->fd == -1)
+	packfd = imsg_get_fd(imsg);
+	if (packfd == -1)
 		return got_error(GOT_ERR_PRIVSEP_NO_FD);
 
-	packfd = imsg->fd;
 	while (!pack_done) {
 		ssize_t r = 0;
 
@@ -1258,7 +1266,10 @@ serve_write(int infd, int outfd, int gotd_sock, const char *repo_path,
 	struct imsg imsg;
 	int report_status = 0;
 
-	imsg_init(&ibuf, gotd_sock);
+	if (imsgbuf_init(&ibuf, gotd_sock) == -1)
+		return got_error_from_errno("imsgbuf_init");
+	imsgbuf_allow_fdpass(&ibuf);
+
 	memset(&imsg, 0, sizeof(imsg));
 
 	err = announce_refs(outfd, &ibuf, 0, repo_path, chattygot);
@@ -1268,7 +1279,8 @@ serve_write(int infd, int outfd, int gotd_sock, const char *repo_path,
 	while (curstate != STATE_EXPECT_PACKFILE) {
 		int n;
 		buf[0] = '\0';
-		err = got_pkt_readpkt(&n, infd, buf, sizeof(buf), chattygot);
+		err = got_pkt_readpkt(&n, infd, buf, sizeof(buf), chattygot,
+		    timeout);
 		if (err)
 			goto done;
 		if (n == 0) {
@@ -1375,7 +1387,7 @@ serve_write(int infd, int outfd, int gotd_sock, const char *repo_path,
 		imsg_free(&imsg);
 	}
 done:
-	imsg_clear(&ibuf);
+	imsgbuf_clear(&ibuf);
 	if (err)
 		echo_error(err, outfd, chattygot);
 	return err;
