@@ -19,6 +19,7 @@
 #include <sys/uio.h>
 #include <sys/time.h>
 
+#include <err.h>
 #include <stdint.h>
 #include <limits.h>
 #include <signal.h>
@@ -34,12 +35,12 @@
 #include "got_object.h"
 
 #include "got_lib_delta.h"
+#include "got_lib_hash.h"
 #include "got_lib_inflate.h"
 #include "got_lib_object.h"
 #include "got_lib_object_parse.h"
 #include "got_lib_object_qid.h"
 #include "got_lib_privsep.h"
-#include "got_lib_hash.h"
 
 static volatile sig_atomic_t sigint_received;
 
@@ -58,13 +59,18 @@ main(int argc, char *argv[])
 
 	signal(SIGINT, catch_sigint);
 
-	imsg_init(&ibuf, GOT_IMSG_FD_CHILD);
+	if (imsgbuf_init(&ibuf, GOT_IMSG_FD_CHILD) == -1) {
+		warn("imsgbuf_init");
+		return 1;
+	}
+	imsgbuf_allow_fdpass(&ibuf);
 
 #ifndef PROFILE
 	/* revoke access to most system calls */
 	if (pledge("stdio recvfd", NULL) == -1) {
 		err = got_error_from_errno("pledge");
 		got_privsep_send_error(&ibuf, err);
+		imsgbuf_clear(&ibuf);
 		return 1;
 	}
 
@@ -134,13 +140,13 @@ done:
 			break;
 	}
 
-	imsg_clear(&ibuf);
 	if (err) {
 		if (!sigint_received && err->code != GOT_ERR_PRIVSEP_PIPE) {
 			fprintf(stderr, "%s: %s\n", getprogname(), err->msg);
 			got_privsep_send_error(&ibuf, err);
 		}
 	}
+	imsgbuf_clear(&ibuf);
 	if (close(GOT_IMSG_FD_CHILD) == -1 && err == NULL)
 		err = got_error_from_errno("close");
 	return err ? 1 : 0;
