@@ -19,6 +19,7 @@
 #include <sys/uio.h>
 #include <sys/time.h>
 
+#include <err.h>
 #include <stdint.h>
 #include <limits.h>
 #include <signal.h>
@@ -35,6 +36,7 @@
 #include "got_repository.h"
 
 #include "got_lib_delta.h"
+#include "got_lib_hash.h"
 #include "got_lib_object.h"
 #include "got_lib_privsep.h"
 #include "got_lib_gitconfig.h"
@@ -138,7 +140,6 @@ send_gitconfig_remotes(struct imsgbuf *ibuf, struct got_remote_repo *remotes,
 		return got_error_from_errno("imsg_compose GITCONFIG_REMOTES");
 
 	err = got_privsep_flush_imsg(ibuf);
-	imsg_clear(ibuf);
 	if (err)
 		return err;
 
@@ -345,13 +346,18 @@ main(int argc, char *argv[])
 #endif
 	signal(SIGINT, catch_sigint);
 
-	imsg_init(&ibuf, GOT_IMSG_FD_CHILD);
+	if (imsgbuf_init(&ibuf, GOT_IMSG_FD_CHILD) == -1) {
+		warn("imsgbuf_init");
+		return 1;
+	}
+	imsgbuf_allow_fdpass(&ibuf);
 
 #ifndef PROFILE
 	/* revoke access to most system calls */
 	if (pledge("stdio recvfd", NULL) == -1) {
 		err = got_error_from_errno("pledge");
 		got_privsep_send_error(&ibuf, err);
+		imsgbuf_clear(&ibuf);
 		return 1;
 	}
 
@@ -442,13 +448,15 @@ main(int argc, char *argv[])
 			break;
 	}
 
-	imsg_clear(&ibuf);
 	if (err) {
 		if (!sigint_received && err->code != GOT_ERR_PRIVSEP_PIPE) {
 			fprintf(stderr, "%s: %s\n", getprogname(), err->msg);
 			got_privsep_send_error(&ibuf, err);
 		}
 	}
+	if (gitconfig)
+		got_gitconfig_close(gitconfig);
+	imsgbuf_clear(&ibuf);
 	if (close(GOT_IMSG_FD_CHILD) == -1 && err == NULL)
 		err = got_error_from_errno("close");
 	return err ? 1 : 0;
