@@ -21,6 +21,7 @@
 #include <sys/time.h>
 #include <sys/stat.h>
 
+#include <err.h>
 #include <stdint.h>
 #include <errno.h>
 #include <limits.h>
@@ -256,14 +257,14 @@ send_ref_status(struct imsgbuf *ibuf, const char *refname, int success,
 		    "unexpected message from server");
 	}
 
-	TAILQ_FOREACH(pe, refs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, refs) {
 		if (strcmp(refname, pe->path) == 0) {
 			ref_valid = 1;
 			break;
 		}
 	}
 	if (!ref_valid) {
-		TAILQ_FOREACH(pe, delete_refs, entry) {
+		RB_FOREACH(pe, got_pathlist_head, delete_refs) {
 			if (strcmp(refname, pe->path) == 0) {
 				ref_valid = 1;
 				break;
@@ -355,9 +356,9 @@ send_pack(int fd, struct got_pathlist_head *refs,
 	struct got_pathlist_entry *pe;
 	int sent_my_capabilites = 0;
 
-	TAILQ_INIT(&their_refs);
+	RB_INIT(&their_refs);
 
-	if (TAILQ_EMPTY(refs) && TAILQ_EMPTY(delete_refs))
+	if (RB_EMPTY(refs) && RB_EMPTY(delete_refs))
 		return got_error(GOT_ERR_SEND_EMPTY);
 
 	while (1) {
@@ -420,9 +421,14 @@ send_pack(int fd, struct got_pathlist_head *refs,
 		if (err)
 			goto done;
 
-		err = got_pathlist_append(&their_refs, refname, id);
-		if (err)
-			goto done;
+		err = got_pathlist_insert(&pe, &their_refs, refname, id);
+		if (err || pe == NULL) {
+			free(refname);
+			free(id);
+			if (err)
+				goto done;
+		}
+
 
 		if (chattygot)
 			fprintf(stderr, "%s: remote has %s %s\n",
@@ -433,7 +439,7 @@ send_pack(int fd, struct got_pathlist_head *refs,
 		id = NULL; /* do not free; owned by their_refs */
 	}
 
-	if (!TAILQ_EMPTY(delete_refs)) {
+	if (!RB_EMPTY(delete_refs)) {
 		if (my_capabilities == NULL ||
 		    strstr(my_capabilities, GOT_CAPA_DELETE_REFS) == NULL) {
 			err = got_error(GOT_ERR_CAPA_DELETE_REFS);
@@ -441,12 +447,12 @@ send_pack(int fd, struct got_pathlist_head *refs,
 		}
 	}
 
-	TAILQ_FOREACH(pe, delete_refs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, delete_refs) {
 		const char *refname = pe->path;
 		struct got_pathlist_entry *their_pe;
 		struct got_object_id *their_id = NULL;
 
-		TAILQ_FOREACH(their_pe, &their_refs, entry) {
+		RB_FOREACH(their_pe, got_pathlist_head, &their_refs) {
 			const char *their_refname = their_pe->path;
 			if (got_path_cmp(refname, their_refname,
 			    strlen(refname), strlen(their_refname)) == 0) {
@@ -480,13 +486,13 @@ send_pack(int fd, struct got_pathlist_head *refs,
 		nsent++;
 	}
 
-	TAILQ_FOREACH(pe, refs, entry) {
+	RB_FOREACH(pe, got_pathlist_head, refs) {
 		const char *refname = pe->path;
 		struct got_object_id *id = pe->data;
 		struct got_object_id *their_id = NULL;
 		struct got_pathlist_entry *their_pe;
 
-		TAILQ_FOREACH(their_pe, &their_refs, entry) {
+		RB_FOREACH(their_pe, got_pathlist_head, &their_refs) {
 			const char *their_refname = their_pe->path;
 			if (got_path_cmp(refname, their_refname,
 			    strlen(refname), strlen(their_refname)) == 0) {
@@ -618,15 +624,20 @@ main(int argc, char **argv)
 		sleep (1);
 #endif
 
-	TAILQ_INIT(&refs);
-	TAILQ_INIT(&delete_refs);
+	RB_INIT(&refs);
+	RB_INIT(&delete_refs);
 
-	imsg_init(&ibuf, GOT_IMSG_FD_CHILD);
+	if (imsgbuf_init(&ibuf, GOT_IMSG_FD_CHILD) == -1) {
+		warn("imsgbuf_init");
+		return 1;
+	}
+	imsgbuf_allow_fdpass(&ibuf);
 #ifndef PROFILE
 	/* revoke access to most system calls */
 	if (pledge("stdio recvfd", NULL) == -1) {
 		err = got_error_from_errno("pledge");
 		got_privsep_send_error(&ibuf, err);
+		imsgbuf_clear(&ibuf);
 		return 1;
 	}
 
@@ -668,6 +679,7 @@ main(int argc, char **argv)
 	for (i = 0; i < send_req.nrefs; i++) {
 		struct got_object_id *id;
 		char *refname;
+		struct got_pathlist_entry *new;
 
 		if ((err = got_privsep_recv_imsg(&imsg, &ibuf, 0)) != 0) {
 			if (err->code == GOT_ERR_PRIVSEP_PIPE)
@@ -717,13 +729,14 @@ main(int argc, char **argv)
 		}
 		memcpy(id, &href.id, sizeof(*id));
 		if (href.delete)
-			err = got_pathlist_append(&delete_refs, refname, id);
+			err = got_pathlist_insert(&new, &delete_refs, refname, id);
 		else
-			err = got_pathlist_append(&refs, refname, id);
-		if (err) {
+			err = got_pathlist_insert(&new, &refs, refname, id);
+		if (err || new == NULL ) {
 			free(refname);
 			free(id);
-			goto done;
+			if (err)
+				goto done;
 		}
 
 		imsg_free(&imsg);
@@ -739,6 +752,7 @@ done:
 		fprintf(stderr, "%s: %s\n", getprogname(), err->msg);
 		got_privsep_send_error(&ibuf, err);
 	}
+	imsgbuf_clear(&ibuf);
 
 	exit(0);
 }
