@@ -32,14 +32,14 @@ SCREENDUMP
 EOF
 
 	cat <<EOF >$testroot/view.expected
-commit $head_id [1/1] master
+$(trim 80 "commit $head_id [1/1] master")
 $ymd flan_hacker  [master] adding the test tree
 
 
 
 
 --------------------------------------------------------------------------------
-[1/40] diff /dev/null $head_id
+$(trim 80 "[1/40] diff /dev/null $head_id")
 commit $head_id (master)
 from: Flan Hacker <flan_hacker@openbsd.org>
 date: $date
@@ -88,7 +88,7 @@ SCREENDUMP
 EOF
 
 	cat <<EOF >$testroot/view.expected
-commit $head_id [1/1] master |[1/40] diff /dev/null $head_id
+$(trim 61 "commit $head_id [1/1] master ")|$(trim 80 "[1/40] diff /dev/null $head_id")
 $ymd flan_hacker  [master] adding the test tree        |commit $head_id (master)
                                                              |from: Flan Hacker <flan_hacker@openbsd.org>
                                                              |date: $date
@@ -105,13 +105,13 @@ $ymd flan_hacker  [master] adding the test tree        |commit $head_id (master)
                                                              |commit - /dev/null
                                                              |commit + $head_id
                                                              |blob - /dev/null
-                                                             |blob + $blobid_alpha (mode 644)
+                                                             |$(trim 80 "blob + $blobid_alpha (mode 644)")
                                                              |--- /dev/null
                                                              |+++ alpha
                                                              |@@ -0,0 +1 @@
                                                              |+alpha
                                                              |blob - /dev/null
-                                                             |blob + $blobid_beta (mode 644)
+                                                             |$(trim 80 "blob + $blobid_beta (mode 644)")
 EOF
 
 	cd $testroot/repo && tog log
@@ -135,7 +135,7 @@ test_log_show_author()
 	local author_time=`git_show_author_time $testroot/repo`
 	local date=`date -u -r $author_time +"%a %b %e %X %Y UTC"`
 	local ymd=`date -u -r $author_time +"%F"`
-	local head_id_len8=`trim_obj_id 32 $head_id`
+	local head_id_len8=`trim_obj_id 8 $head_id`
 
 	echo "mod alpha" > $testroot/repo/alpha
 	cd $testroot/repo && git add .
@@ -144,7 +144,7 @@ test_log_show_author()
 	    /dev/null
 
 	local commit1=`git_show_head $testroot/repo`
-	local id1_len8=`trim_obj_id 32 $commit1`
+	local id1_len8=`trim_obj_id 8 $commit1`
 
 	cat <<EOF >$TOG_TEST_SCRIPT
 @		toggle show author
@@ -193,7 +193,7 @@ SCREENDUMP
 EOF
 
 	cat <<EOF >$testroot/view.expected
-commit $commit1 [1/2] master
+$(trim 80 "commit $commit1 [1/2] master")
 $ymd flan_hacker  $scrolled_msg
 $ymd flan_hacker  ng the test tree
 EOF
@@ -227,7 +227,7 @@ SCREENDUMP
 EOF
 
 	cat <<EOF >$testroot/view.expected
-commit $head_id [1/1] master
+$(trim 80 "commit $head_id [1/1] master")
 $ymd flan_hacker  [master] adding the test tree
 
 --------------------------------------------------------------------------------
@@ -269,7 +269,7 @@ SCREENDUMP
 EOF
 
 	cat <<EOF >$testroot/view.expected
-commit $head_id [1/1] master
+$(trim 80 "commit $head_id [1/1] master")
 $ymd flan_hacker  [master] adding the test tree
 
 --------------------------------------------------------------------------------
@@ -290,6 +290,49 @@ EOF
 		return 1
 	fi
 
+	# test T keymap on a work tree entry
+	got checkout "$testroot/repo" "$testroot/wt" > /dev/null
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		echo "got checkout failed unexpectedly"
+		test_done "$testroot" "$ret"
+		return 1
+	fi
+
+	cd "$testroot/wt"
+	echo "'alpha" >> alpha
+
+	local id10=$(trim_obj_id 10 $head_id)
+
+	cat <<-EOF >$TOG_TEST_SCRIPT
+	WAIT_FOR_UI	wait for log thread to fetch wt state
+	T		open tree view
+	S		toggle horizontal split
+	SCREENDUMP
+	EOF
+
+	cat <<EOF >$testroot/view.expected
+diff $testroot/wt (work tree changes) [0/1] maste
+$ymd flan_hacker  work tree changes based on [$id10]
+--------------------------------------------------------------------------------
+commit $head_id
+[1/4] /
+
+  alpha
+  beta
+  epsilon/
+  gamma/
+EOF
+
+	tog log
+	cmp -s "$testroot/view.expected" "$testroot/view"
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u "$testroot/view.expected" "$testroot/view"
+		test_done "$testroot" "$ret"
+		return 1
+	fi
+
 	test_done "$testroot" "$ret"
 }
 
@@ -399,11 +442,11 @@ test_log_commit_keywords()
 
 	cat <<-EOF >$testroot/view.expected
 	commit $(pop_idx 5 $@) [1/5]
-	$ymd $(trim_obj_id 32 $(pop_idx 5 $@)) flan_hacker  commit 4
-	$ymd $(trim_obj_id 32 $(pop_idx 4 $@)) flan_hacker  commit 3
-	$ymd $(trim_obj_id 32 $(pop_idx 3 $@)) flan_hacker  commit 2
-	$ymd $(trim_obj_id 32 $(pop_idx 2 $@)) flan_hacker  commit 1
-	$ymd $(trim_obj_id 32 $(pop_idx 1 $@)) flan_hacker  adding the test tree
+	$ymd $(trim_obj_id 8 $(pop_idx 5 $@)) flan_hacker  commit 4
+	$ymd $(trim_obj_id 8 $(pop_idx 4 $@)) flan_hacker  commit 3
+	$ymd $(trim_obj_id 8 $(pop_idx 3 $@)) flan_hacker  commit 2
+	$ymd $(trim_obj_id 8 $(pop_idx 2 $@)) flan_hacker  commit 1
+	$ymd $(trim_obj_id 8 $(pop_idx 1 $@)) flan_hacker  adding the test tree
 
 
 
@@ -421,13 +464,13 @@ test_log_commit_keywords()
 
 	cat <<-EOF >$testroot/view.expected
 	commit $(pop_idx 7 $@) [1/7]
-	$ymd $(trim_obj_id 32 $(pop_idx 7 $@)) flan_hacker  commit 6
-	$ymd $(trim_obj_id 32 $(pop_idx 6 $@)) flan_hacker  commit 5
-	$ymd $(trim_obj_id 32 $(pop_idx 5 $@)) flan_hacker  commit 4
-	$ymd $(trim_obj_id 32 $(pop_idx 4 $@)) flan_hacker  commit 3
-	$ymd $(trim_obj_id 32 $(pop_idx 3 $@)) flan_hacker  commit 2
-	$ymd $(trim_obj_id 32 $(pop_idx 2 $@)) flan_hacker  commit 1
-	$ymd $(trim_obj_id 32 $(pop_idx 1 $@)) flan_hacker  adding the test tree
+	$ymd $(trim_obj_id 8 $(pop_idx 7 $@)) flan_hacker  commit 6
+	$ymd $(trim_obj_id 8 $(pop_idx 6 $@)) flan_hacker  commit 5
+	$ymd $(trim_obj_id 8 $(pop_idx 5 $@)) flan_hacker  commit 4
+	$ymd $(trim_obj_id 8 $(pop_idx 4 $@)) flan_hacker  commit 3
+	$ymd $(trim_obj_id 8 $(pop_idx 3 $@)) flan_hacker  commit 2
+	$ymd $(trim_obj_id 8 $(pop_idx 2 $@)) flan_hacker  commit 1
+	$ymd $(trim_obj_id 8 $(pop_idx 1 $@)) flan_hacker  adding the test tree
 
 
 	EOF
@@ -443,11 +486,11 @@ test_log_commit_keywords()
 
 	cat <<-EOF >$testroot/view.expected
 	commit $(pop_idx 5 $@) [1/5]
-	$ymd $(trim_obj_id 32 $(pop_idx 5 $@)) flan_hacker  commit 4
-	$ymd $(trim_obj_id 32 $(pop_idx 4 $@)) flan_hacker  commit 3
-	$ymd $(trim_obj_id 32 $(pop_idx 3 $@)) flan_hacker ~commit 2
-	$ymd $(trim_obj_id 32 $(pop_idx 2 $@)) flan_hacker  commit 1
-	$ymd $(trim_obj_id 32 $(pop_idx 1 $@)) flan_hacker  adding the test tree
+	$ymd $(trim_obj_id 8 $(pop_idx 5 $@)) flan_hacker  commit 4
+	$ymd $(trim_obj_id 8 $(pop_idx 4 $@)) flan_hacker  commit 3
+	$ymd $(trim_obj_id 8 $(pop_idx 3 $@)) flan_hacker ~commit 2
+	$ymd $(trim_obj_id 8 $(pop_idx 2 $@)) flan_hacker  commit 1
+	$ymd $(trim_obj_id 8 $(pop_idx 1 $@)) flan_hacker  adding the test tree
 
 
 
@@ -473,7 +516,7 @@ test_log_commit_keywords()
 
 	cat <<-EOF >$testroot/view.expected
 	commit $(pop_idx 1 $@) [1/1]
-	$ymd $(trim_obj_id 32 $(pop_idx 1 $@)) flan_hacker  adding the test tree
+	$ymd $(trim_obj_id 8 $(pop_idx 1 $@)) flan_hacker  adding the test tree
 
 
 
@@ -499,7 +542,7 @@ test_log_commit_keywords()
 test_log_show_base_commit()
 {
 	# make view wide enough to show full headline
-	test_init log_show_base_commit 80 3
+	test_init log_show_base_commit 80 4
 	local repo="$testroot/repo"
 	local id=$(git_show_head "$repo")
 
@@ -528,9 +571,10 @@ test_log_show_base_commit()
 	EOF
 
 	cat <<-EOF >$testroot/view.expected
-	commit $head_id [1/2] master
+	$(trim 80 "commit $head_id [1/2] master")
 	$ymd flan_hacker *[master] base commit
 	$ymd flan_hacker  adding the test tree
+
 	EOF
 
 	tog log
@@ -544,9 +588,10 @@ test_log_show_base_commit()
 
 	# check marker is not drawn when not in a work tree
 	cat <<-EOF >$testroot/view.expected
-	commit $head_id [1/2] master
+	$(trim 80 "commit $head_id [1/2] master")
 	$ymd flan_hacker  [master] base commit
 	$ymd flan_hacker  adding the test tree
+
 	EOF
 
 	tog log -r "$repo"
@@ -569,9 +614,34 @@ test_log_show_base_commit()
 	EOF
 
 	cat <<-EOF >$testroot/view.expected
-	commit $head_id [1/3] master
+	$(trim 80 "commit $head_id [1/3] master")
 	$ymd flan_hacker ~[master] new base mixed-commit
 	$ymd flan_hacker  base commit
+	$ymd flan_hacker  adding the test tree
+	EOF
+
+	tog log
+	cmp -s "$testroot/view.expected" "$testroot/view"
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u "$testroot/view.expected" "$testroot/view"
+		test_done "$testroot" "$ret"
+		return 1
+	fi
+
+	# check marker is not incorrectly drawn when opening a nested log view
+	cat <<-EOF >$TOG_TEST_SCRIPT
+	T		# open tree view of base commit
+	j		# select beta tree entry
+	L		# load log view of commits involving beta
+	SCREENDUMP
+	EOF
+
+	cat <<-EOF >$testroot/view.expected
+	$(trim 80 "commit $id /beta [1/1]")
+	$ymd flan_hacker  adding the test tree
+
+
 	EOF
 
 	tog log
@@ -648,7 +718,7 @@ test_log_limit_view()
 	EOF
 
 	cat <<-EOF >$testroot/view.expected
-	commit $id [1/2] master
+	$(trim 80 "commit $id [1/2] master")
 	$ymd flan_hacker  [master] beta1
 	$ymd flan_hacker  beta0
 
@@ -707,7 +777,7 @@ test_log_search()
 	EOF
 
 	cat <<-EOF >$testroot/view.expected
-	commit $(pop_idx 9 $@) [9/17] no more matches
+	$(trim 80 "commit $(pop_idx 9 $@) [9/17] no more matches")
 	$ymd flan_hacker  alpha commit 14
 	$ymd flan_hacker  alpha commit 13
 	$ymd flan_hacker  alpha commit 12
@@ -726,6 +796,344 @@ test_log_search()
 	test_done "$testroot" "$ret"
 }
 
+test_log_mark_keymap()
+{
+	test_init log_mark_keymap 141 14
+
+	local repo="$testroot/repo"
+	local wt="$testroot/wt"
+	local id_root=$(git_show_head $repo)
+	local prefix_root=$(trim_obj_id 8 $id_root)
+	local author_time=$(git_show_author_time $repo)
+	local ymd_root=$(date -u -r $author_time +"%F")
+	local alpha_root=$(get_blob_id $testroot/repo "" alpha)
+
+	got checkout "$repo" "$wt" > /dev/null
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		echo "got checkout failed unexpectedly"
+		test_done "$testroot" "$ret"
+		return 1
+	fi
+
+	cd "$wt"
+
+	echo "new alpha" > alpha
+	got commit -m "new alpha" > /dev/null
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		echo "got commit failed unexpectedly" >&2
+		test_done "$testroot" $ret
+		return 1
+	fi
+
+	author_time=$(git_show_author_time $repo)
+	local ymd_head=$(date -u -r $author_time +"%F")
+	local id_head=$(git_show_head $repo)
+	local prefix_head=$(trim_obj_id 8 $id_head)
+	local alpha_head=$(get_blob_id $testroot/repo "" alpha)
+
+	# test marker is correctly applied to arbitrary commit
+	cat <<-EOF >$TOG_TEST_SCRIPT
+	j
+	m		mark commit
+	WAIT_FOR_UI
+	SCREENDUMP
+	EOF
+
+	cat <<-EOF >$testroot/view.expected
+	commit $id_root [2/2]
+	$ymd_head $prefix_head flan_hacker ~[master] new alpha
+	$ymd_root $prefix_root flan_hacker >adding the test tree
+
+
+
+
+
+
+
+
+
+
+
+	EOF
+
+	tog log
+	cmp -s $testroot/view.expected $testroot/view
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u $testroot/view.expected $testroot/view
+		test_done "$testroot" $ret
+		return 1
+	fi
+
+	# test commit is correctly unmarked
+	cat <<-EOF >$TOG_TEST_SCRIPT
+	j
+	m		mark commit
+	m		unmark commit
+	WAIT_FOR_UI
+	SCREENDUMP
+	EOF
+
+	cat <<-EOF >$testroot/view.expected
+	commit $id_root [2/2]
+	$ymd_head $prefix_head flan_hacker ~[master] new alpha
+	$ymd_root $prefix_root flan_hacker  adding the test tree
+
+
+
+
+
+
+
+
+
+
+
+	EOF
+
+	tog log
+	cmp -s $testroot/view.expected $testroot/view
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u $testroot/view.expected $testroot/view
+		test_done "$testroot" $ret
+		return 1
+	fi
+
+	# test marker correctly overwrites base commit marker
+	cat <<-EOF >$TOG_TEST_SCRIPT
+	m		mark commit
+	SCREENDUMP
+	EOF
+
+	cat <<-EOF >$testroot/view.expected
+	commit $id_head [1/2] master
+	$ymd_head $prefix_head flan_hacker >[master] new alpha
+	$ymd_root $prefix_root flan_hacker  adding the test tree
+
+
+
+
+
+
+
+
+
+
+
+	EOF
+
+	tog log
+	cmp -s $testroot/view.expected $testroot/view
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u $testroot/view.expected $testroot/view
+		test_done "$testroot" $ret
+		return 1
+	fi
+
+	# test diff of marked and selected commit is correctly rendered
+	cat <<-EOF >$TOG_TEST_SCRIPT
+	m		mark commit
+	j
+	KEY_ENTER	show diff of marked and root commit
+	F		toggle fullscreen
+	SCREENDUMP
+	EOF
+
+	cat <<-EOF >$testroot/view.expected
+	[1/14] diff $id_head $id_root
+	M  alpha  |  1+  1-
+
+	1 file changed, 1 insertion(+), 1 deletion(-)
+
+	commit - $id_head
+	commit + $id_root
+	blob - $alpha_head
+	blob + $alpha_root
+	--- alpha
+	+++ alpha
+	@@ -1 +1 @@
+	-new alpha
+	+alpha
+	EOF
+
+	tog log
+	cmp -s $testroot/view.expected $testroot/view
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u $testroot/view.expected $testroot/view
+		test_done "$testroot" $ret
+		return 1
+	fi
+
+	test_done "$testroot" $ret
+}
+
+test_log_worktree_entries()
+{
+	test_init log_worktree_entries 96 4
+	local repo="$testroot/repo"
+
+	got checkout "$repo" "$testroot/wt" > /dev/null
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		echo "got checkout failed unexpectedly"
+		test_done "$testroot" "$ret"
+		return 1
+	fi
+
+	# test dirty work tree (no staged changes)
+	cd "$testroot/wt"
+	echo "'alpha" >> alpha
+
+	local id=$(git_show_head "$repo")
+	local id10=$(trim_obj_id 10 $id)
+	local author_time=$(git_show_author_time "$repo")
+	local ymd=$(date -u -r "$author_time" +"%F")
+
+	cat <<-EOF >$TOG_TEST_SCRIPT
+	WAIT_FOR_UI	wait for log thread to fetch wt state
+	WAIT_FOR_UI	trigger redraw to reveal work tree entries
+	SCREENDUMP
+	EOF
+
+	cat <<-EOF >$testroot/view.expected
+	diff $testroot/wt (work tree changes) [0/1] master
+	$ymd flan_hacker  work tree changes based on [$id10]
+	$ymd flan_hacker *[master] adding the test tree
+
+	EOF
+
+	tog log
+	cmp -s "$testroot/view.expected" "$testroot/view"
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u "$testroot/view.expected" "$testroot/view"
+		test_done "$testroot" "$ret"
+		return 1
+	fi
+
+	# test dirty work tree with staged changes too
+	echo "'beta" >> beta
+	got stage beta > /dev/null
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		echo "got stage failed unexpectedly"
+		test_done "$testroot" "$ret"
+		return 1
+	fi
+
+	cat <<-EOF >$TOG_TEST_SCRIPT
+	WAIT_FOR_UI	wait for log thread to fetch wt state
+	j		select staged entry
+	SCREENDUMP
+	EOF
+
+	cat <<-EOF >$testroot/view.expected
+	diff -s $testroot/wt (staged work tree changes) [0/1] master
+	$ymd flan_hacker  work tree changes based on [$id10]
+	$ymd flan_hacker  staged work tree changes based on [$id10]
+	$ymd flan_hacker *[master] adding the test tree
+	EOF
+
+	tog log
+	cmp -s "$testroot/view.expected" "$testroot/view"
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u "$testroot/view.expected" "$testroot/view"
+		test_done "$testroot" "$ret"
+		return 1
+	fi
+
+	# test with only staged changes
+	got revert alpha > /dev/null
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		echo "got revert failed unexpectedly"
+		test_done "$testroot" "$ret"
+		return 1
+	fi
+
+	cat <<-EOF >$TOG_TEST_SCRIPT
+	WAIT_FOR_UI	wait for log thread to fetch wt state
+	WAIT_FOR_UI	trigger redraw to reveal work tree entries
+	SCREENDUMP
+	EOF
+
+	cat <<-EOF >$testroot/view.expected
+	diff -s $testroot/wt (staged work tree changes) [0/1] master
+	$ymd flan_hacker  staged work tree changes based on [$id10]
+	$ymd flan_hacker *[master] adding the test tree
+
+	EOF
+
+	tog log
+	cmp -s "$testroot/view.expected" "$testroot/view"
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u "$testroot/view.expected" "$testroot/view"
+		test_done "$testroot" "$ret"
+		return 1
+	fi
+
+	# test with no author set
+	echo "'alpha" >> alpha
+
+	cat <<-EOF >$testroot/view.expected
+	diff $testroot/wt (work tree changes) [0/1] master
+	$ymd              work tree changes based on [$id10]
+	$ymd              staged work tree changes based on [$id10]
+	$ymd flan_hacker *[master] adding the test tree
+	EOF
+
+	unset GOT_AUTHOR
+	tog log
+	cmp -s "$testroot/view.expected" "$testroot/view"
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u "$testroot/view.expected" "$testroot/view"
+		test_done "$testroot" "$ret"
+		return 1
+	fi
+
+	test_done "$testroot" 0
+}
+
+test_log_tiny_child_tree_view()
+{
+	test_init log_tiny_child_tree_view 120 3
+
+	local id=$(git_show_head $testroot/repo)
+
+	# This test covers the offset_selection_down() path ensuring
+	# indexes are properly clamped to prevent negative values.
+	cat <<-EOF >$TOG_TEST_SCRIPT
+	T		open tree view in vsplit
+	F		toggle fullscreen to hit offset_selection_down()
+	SCREENDUMP
+	EOF
+
+	cat <<-EOF >$testroot/view.expected
+	commit $id
+	[1/4] /
+
+	EOF
+
+	cd $testroot/repo && tog log
+	cmp -s $testroot/view.expected $testroot/view
+	ret=$?
+	if [ $ret -ne 0 ]; then
+		diff -u $testroot/view.expected $testroot/view
+		test_done "$testroot" "$ret"
+		return 1
+	fi
+
+	test_done "$testroot" "$ret"
+}
+
 test_parseargs "$@"
 run_test test_log_hsplit_diff
 run_test test_log_vsplit_diff
@@ -738,3 +1146,6 @@ run_test test_log_commit_keywords
 run_test test_log_show_base_commit
 run_test test_log_limit_view
 run_test test_log_search
+run_test test_log_mark_keymap
+run_test test_log_worktree_entries
+run_test test_log_tiny_child_tree_view
